@@ -1,6 +1,6 @@
 ;;;; batch-driver.scm - Driver procedure for the compiler
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -44,7 +44,7 @@
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   compiler-cleanup-hook check-global-exports disabled-warnings check-global-imports
-  file-io-only
+  file-io-only undefine-shadowed-macros
   unit-name insert-timer-checks used-units inline-max-size
   debugging perform-lambda-lifting! disable-stack-overflow-checking
   foreign-declarations emit-trace-info block-compilation line-number-database-size
@@ -84,6 +84,7 @@
 
 (define-constant default-profile-name "PROFILE")
 (define-constant default-inline-max-size 10)
+(define-constant funny-message-timeout 60000)
 
 
 ;;; Compile a complete source file:
@@ -128,11 +129,13 @@
 	[hshrink (memq 'heap-shrinkage options)]
 	[kwstyle (memq 'keyword-style options)]
 	[lcthreshold (memq 'compress-literals options)]
+	[uses-units '()]
 	[uunit (memq 'unit options)]
 	[a-only (memq 'analyze-only options)]
 	[dynamic (memq 'dynamic options)]
 	[dumpnodes #f]
 	[quiet (memq 'quiet options)]
+	[start-time #f]
 	[ssize (or (memq 'nursery options) (memq 'stack-size options))] )
 
     (define (cputime) (##sys#fudge 6))
@@ -265,6 +268,9 @@
 	      ipath) )
     (when (and outfile filename (string=? outfile filename))
       (quit "source- and output-filename are the same") )
+    (set! uses-units (map string->symbol (collect-options 'uses)))
+    (when (memq 'keep-shadowed-macros options)
+      (set! undefine-shadowed-macros #f) )
 
     ;; Handle feature options:
     (for-each register-feature! (collect-options 'feature))
@@ -297,7 +303,7 @@
 		     (else (quit "no filename available for `-extension' option")) ) ) ) ) ) ) )
 
     ;; Append required extensions to initforms:
-    (let ([ids (map string->symbol (collect-options 'require-extension))])
+    (let ([ids (lset-difference eq? (map string->symbol (collect-options 'require-extension)) uses-units)])
       (set! initforms
 	(append initforms (map (lambda (r) `(##core#require-extension ',r)) ids)) ) )
 
@@ -364,6 +370,7 @@
 	   (debugging 'r "debugging options" debugging-chicken)
 	   (debugging 'r "target heap size" target-heap-size)
 	   (debugging 'r "target stack size" target-stack-size)
+	   (set! start-time (cputime))
 
 	   ;; Read toplevel expressions:
 	   (set! ##sys#line-number-database (make-vector line-number-database-size '()))
@@ -397,13 +404,13 @@
 	   (let ([proc (user-preprocessor-pass)])
 	     (when proc
 	       (when verbose (printf "User preprocessing pass...~%~!"))
-	       (set! forms (cons (first forms) (map proc (cdr forms)))) ) )
+	       (set! forms (map proc forms))))
 
 	   (print "source" '|1| forms)
 	   (begin-time)
-	   (let ([us (collect-options 'uses)])
-	     (unless (null? us)
-	       (set! forms (cons `(declare (uses ,@us)) forms)) ) )
+	   (unless (null? uses-units)
+	     (set! ##sys#explicit-library-modules (append ##sys#explicit-library-modules uses-units))
+	     (set! forms (cons `(declare (uses ,@uses-units)) forms)) )
 	   (let* ([exps0 (map canonicalize-expression (append initforms forms))]
 		  [pvec (gensym)]
 		  [plen (length profile-lambda-list)]
@@ -557,6 +564,8 @@
 			    (let ([node3 (perform-closure-conversion node2 db)])
 			      (end-time "closure conversion")
 			      (print-db "final-analysis" '|8| db i)
+			      (when (and ##sys#warnings-enabled (> (- (cputime) start-time) funny-message-timeout))
+				(display "(don't despair - still compiling...)\n") )
 			      (when export-file-name
 				(dump-exported-globals db export-file-name) )
 			      (let ([upap (user-post-analysis-pass)])
