@@ -1,6 +1,6 @@
 ;;;; c-platform.scm - Platform specific parameters and definitions
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -122,7 +122,8 @@
 
 (define valid-compiler-options
   '(-help h help version verbose explicit-use quiet no-trace no-warnings unsafe block
-    check-syntax to-stdout no-usual-integrations case-insensitive no-lambda-info profile inline
+    check-syntax to-stdout no-usual-integrations case-insensitive no-lambda-info 
+    profile inline keep-shadowed-macros
     fixnum-arithmetic disable-interrupts optimize-leaf-routines check-imports
     lambda-lift run-time-macros tag-pointers accumulate-profile
     disable-stack-overflow-checks disable-c-syntax-checks unsafe-libraries raw 
@@ -161,14 +162,14 @@
 (define default-extended-bindings
   '(bitwise-and bitwise-ior bitwise-xor bitwise-not add1 sub1 fx+ fx- fx* fx/ fxmod
     fx= fx> fx< fx>= fx<= fixnum? fxneg fxmax fxmin identity fp+ fp- fp* fp/ fpmin fpmax fpneg
-    fp> fp< fp= fp>= fp<= fxand fxnot fxior fxxor fxshr fxshl
+    fp> fp< fp= fp>= fp<= fxand fxnot fxior fxxor fxshr fxshl bit-set?
     arithmetic-shift void flush-output thread-specific thread-specific-set!
     not-pair? atom? null-list? print print* error cpu-time proper-list? call/cc
     u8vector->byte-vector s8vector->byte-vector u16vector->byte-vector s16vector->byte-vector 
     u32vector->byte-vector 
     s32vector->byte-vector byte-vector-length block-ref block-set! number-of-slots
     f32vector->byte-vector f64vector->byte-vector byte-vector-ref byte-vector-set!
-    hash-table-ref
+    hash-table-ref any?
     first second third fourth make-record-instance
     u8vector-length s8vector-length u16vector-length s16vector-length u32vector-length s32vector-length
     f32vector-length f64vector-length setter
@@ -183,14 +184,14 @@
     ##sys#call-with-current-continuation ##sys#size ##sys#byte ##sys#setbyte
     ##sys#pointer? ##sys#generic-structure? ##sys#structure? ##sys#check-structure
     ##sys#check-exact ##sys#check-number ##sys#check-list ##sys#check-pair ##sys#check-string ##sys#check-symbol 
-    ##sys#check-char ##sys#check-vector ##sys#check-byte-vector ##sys#list
+    ##sys#check-char ##sys#check-vector ##sys#check-byte-vector ##sys#list ##sys#cons
     ##sys#call-with-values ##sys#fits-in-int? ##sys#fits-in-unsigned-int? ##sys#flonum-in-fixnum-range? 
     ##sys#fudge ##sys#immediate? ##sys#direct-return ##sys#context-switch
     ##sys#make-structure ##sys#apply ##sys#apply-values ##sys#continuation-graft
     ##sys#bytevector? ##sys#make-vector ##sys#setter
     ##sys#foreign-char-argument ##sys#foreign-fixnum-argument ##sys#foreign-flonum-argument
     ##sys#foreign-block-argument ##sys#foreign-number-vector-argument
-    ##sys#foreign-string-argument ##sys#foreign-pointer-argument
+    ##sys#foreign-string-argument ##sys#foreign-pointer-argument ##sys#void
     ##sys#foreign-integer-argument ##sys#foreign-unsigned-integer-argument ##sys#double->number
     ##sys#peek-fixnum ##sys#setislot ##sys#poke-integer ##sys#permanent? ##sys#values ##sys#poke-double
     ##sys#intern-symbol ##sys#make-symbol ##sys#null-pointer? ##sys#peek-byte) )
@@ -217,7 +218,7 @@
     f32vector->byte-vector f64vector->byte-vector ##sys#byte ##sys#setbyte byte-vector-ref byte-vector-set!
     u8vector-length s8vector-length u16vector-length s16vector-length u32vector-length s32vector-length
     f32vector-length f64vector-length ##sys#apply-values ##sys#setter setter
-    u8vector-ref s8vector-ref u16vector-ref s16vector-ref
+    u8vector-ref s8vector-ref u16vector-ref s16vector-ref u32vector-ref s32vector-ref
     u8vector-set! s8vector-set! u16vector-set! s16vector-set! u32vector-set! s32vector-set!
     ##sys#intern-symbol ##sys#make-symbol make-record-instance error cpu-time ##sys#block-set!) )
 
@@ -906,6 +907,9 @@
 (rewrite 'u16vector-ref 2 2 "C_u_i_u16vector_ref" #f #f)
 (rewrite 's16vector-ref 2 2 "C_u_i_s16vector_ref" #f #f)
 
+(rewrite 'u32vector-ref 22 2 "C_a_i_u32vector_ref" #f words-per-flonum "C_u_i_u32vector_ref")
+(rewrite 's32vector-ref 22 2 "C_a_i_s32vector_ref" #f words-per-flonum "C_u_i_s32vector_ref")
+
 (rewrite 'u8vector-set! 2 3 "C_u_i_u8vector_set" #f #f)
 (rewrite 's8vector-set! 2 3 "C_u_i_s8vector_set" #f #f)
 (rewrite 'u16vector-set! 2 3 "C_u_i_u16vector_set" #f #f)
@@ -1033,3 +1037,32 @@
 			 '##core#call '(#t)
 			 (list cont (varnode (cdr a))) ) ) ) ) ) ) ) ) )
 			       
+(rewrite 'void 3 '##sys#undefined-value)
+(rewrite '##sys#void 3 '##sys#undefined-value)
+
+(rewrite
+ 'any? 8
+ (lambda (db classargs cont callargs) 
+   (and (= 1 (length callargs))
+	(let ((arg (car callargs)))
+	  (make-node
+	   '##core#call '(#t) 
+	   (list cont
+		 (if (and (eq? '##core#variable (node-class arg))
+			  (not (get db (car (node-parameters arg)) 'global)) )
+		     (qnode #t)
+		     (make-node 
+		      '##core#inline '("C_anyp")
+		      (list arg)) ) ) ) ) ) ) )
+
+(rewrite
+ 'bit-set? 8
+ (lambda (db classargs cont callargs)
+   (and (= 2 (length callargs))
+	(make-node
+	 '##core#call '(#t)
+	 (list cont
+	       (make-node
+		'##core#inline 
+		(list (if (eq? number-type 'fixnum) "C_u_i_bit_setp" "C_i_bit_setp"))
+		callargs) ) ) ) ) )
