@@ -1,6 +1,6 @@
 ;;;; chicken-setup
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -36,10 +36,14 @@
 (declare
   (run-time-macros)
   (uses srfi-1 regex utils posix tcp match srfi-18 srfi-13)
-  (export move-file run:execute make/proc uninstall-extension install-extension install-program install-script
-	  setup-verbose-flag setup-install-flag installation-prefix find-library find-header
-	  program-path remove-file* patch yes-or-no? setup-build-directory setup-root-directory create-directory
-	  test-compile copy-file run-verbose) )
+  (export move-file run:execute make/proc uninstall-extension
+	  install-extension install-program install-script
+	  setup-verbose-flag setup-install-flag installation-prefix
+	  find-library find-header program-path remove-file* patch
+	  yes-or-no? setup-build-directory setup-root-directory
+	  create-directory test-compile try-compile copy-file run-verbose
+	  required-chicken-version required-extension-version
+	  cross-chicken) )
 
 #>
 #ifndef C_USE_C_DEFAULTS
@@ -96,10 +100,10 @@ static void create_directory(char *pathname) {}
 (define-constant long-options
   '("-help" "-uninstall" "-list" "-run" "-repository" "-program-path" "-version" "-script" "-check"
     "-fetch" "-host" "-proxy" "-keep" "-verbose" "-csc-option" "-dont-ask" "-no-install" "-docindex" "-eval"
-    "-debug" "-ls" "-release" "-test") )
+    "-debug" "-ls" "-release" "-test" "-fetch-tree" "-tree" "-svn" "-local" "-destdir" "-revision") )
 
 (define-constant short-options
-  '(#\h #\u #\l #\r #\R #\P #\V #\s #\C #\f #\H #\p #\k #\v #\c #\d #\n #\i #\e #\D #f #f #\t) )
+  '(#\h #\u #\l #\r #\R #\P #\V #\s #\C #\f #\H #\p #\k #\v #\c #\d #\n #\i #\e #\D #f #f #\t #f #f #f #f #f #f) )
 
 
 (define *install-bin-path* 
@@ -127,6 +131,8 @@ static void create_directory(char *pathname) {}
 (define *remove-command* (if *windows-shell* "del /Q /S" "rm -fr"))
 (define *move-command* (if *windows-shell* 'move 'mv))
 
+(define (cross-chicken) (##sys#fudge 39))
+
 (define create-directory
   (let ()
     (define (verb dir)
@@ -145,7 +151,7 @@ static void create_directory(char *pathname) {}
       str) )
 
 (define setup-root-directory (make-parameter #f))
-(define setup-build-directory (make-parameter #f))
+(define setup-build-directory (make-parameter (current-directory)))
 (define setup-verbose-flag (make-parameter #f))
 (define setup-install-flag (make-parameter #t))
 
@@ -154,7 +160,7 @@ static void create_directory(char *pathname) {}
 (define *tmpdir-created* #f)
 (define *keep-stuff* #f)
 (define *csc-options* '())
-(define *abort-yes-or-no* #f)
+(define *abort-hook* #f)
 (define *dont-ask* #f)
 (define *rebuild-doc-index* #f)
 (define *check-repository* #f)
@@ -165,13 +171,16 @@ static void create_directory(char *pathname) {}
 (define *example-directory* (make-pathname (chicken-home) "examples"))
 (define *base-directory* (current-directory))
 (define *fetch-tree-only* #f)
+(define *svn-repository* #f)
+(define *local-repository* #f)
+(define *destdir* #f)
+(define *repository-hosts* '(("www.call-with-current-continuation.org" "eggs" 80)))
+(define *revision* #f)
+(define *repository-tree-downloaded* #f)
 
 
-; Repository-format:
-;
-; ((NAME FILENAME REQUIRED-NAME ...) ...)
-(define repository-hosts '(("www.call-with-current-continuation.org" "eggs" 80)))
-
+(define (abort-setup)
+  (*abort-hook* #f) )
 
 (define (yes-or-no? str . default)
   (let ((def (:optional default #f)))
@@ -184,7 +193,7 @@ static void create_directory(char *pathname) {}
 	      ((and def (string=? "" ln)) (set! ln def)) )
 	(cond ((string-ci=? "yes" ln) #t)
 	      ((string-ci=? "no" ln) #f)
-	      ((string-ci=? "abort" ln) (*abort-yes-or-no* #f))
+	      ((string-ci=? "abort" ln) (abort-setup))
 	      (else
 	       (printf "~%Please enter \"yes\", \"no\" or \"abort\".~%")
 	       (loop) ) ) ) ) ) )
@@ -225,9 +234,6 @@ static void create_directory(char *pathname) {}
       (pathname-replace-extension file ##sys#load-dynamic-extension)
       file) )
 
-(define (unformatify str)
-  (string-translate* str '(("~" . "~~"))) )
-
 (define (run:execute explist)
   (define (smooth lst)
     (let ((slst (map ->string lst)))
@@ -235,7 +241,7 @@ static void create_directory(char *pathname) {}
   (for-each
    (lambda (cmd)
      (when (run-verbose) (printf "  ~A~%~!" cmd))
-     (system* (unformatify cmd) ))
+     (system* "~a" cmd) )
    (map smooth explist) ) )
 
 (define-macro (run . explist)
@@ -285,11 +291,11 @@ static void create_directory(char *pathname) {}
 
 (define (make:check-argv argv)
   (or (string? argv)
-      (and (vector? argv)
-	   (andmap string? (vector->list argv)))
-      (error "argument is not a string or string vector" argv)))
+      (andmap string? argv)
+      (error "argument is not a string or string list" argv)))
 
 (define (make:make/proc/helper spec argv)
+  (when (vector? argv) (set! argv (vector->list argv)))
   (make:check-spec spec)
   (make:check-argv argv)
   (letrec ((made '())
@@ -301,10 +307,8 @@ static void create_directory(char *pathname) {}
 		     (s2 (fixmaketarget s)) 
 		     (date (and (file-exists? s2)
 				(file-modification-time s2))))
-
 		(when (setup-verbose-flag)
 		  (printf "make: ~achecking ~a~%" indent s2))
-
 		(if line
 		    (let ((deps (cadr line)))
 		      (for-each (let ((new-indent (string-append " " indent)))
@@ -348,8 +352,8 @@ static void create_directory(char *pathname) {}
 		      (error (sprintf "don't know how to make ~a" s2))))))))
     (cond
      ((string? argv) (make-file argv ""))
-     ((equal? argv '#()) (make-file (caar spec) ""))
-     (else (for-each (lambda (f) (make-file f "")) (vector->list argv))))
+     ((null? argv) (make-file (caar spec) ""))
+     (else (for-each (lambda (f) (make-file f "")) argv)))
     (when (setup-verbose-flag)
       (for-each (lambda (item)
 		  (printf "make: made ~a~%" item))
@@ -357,35 +361,31 @@ static void create_directory(char *pathname) {}
 
 (define make/proc
   (case-lambda
-   ((spec) (make:make/proc/helper spec '#()))
+   ((spec) (make:make/proc/helper spec '()))
    ((spec argv) (make:make/proc/helper spec argv))))
 
-(define-macro (make spec . argv)
-  (let ((make (lambda (spec argv)
-		(let ((form-error (lambda (s . p) (apply error s spec p))))
-		  (and (or (list? spec) (form-error "illegal specification (not a sequence)"))
-		       (or (pair? spec) (form-error "empty specification"))
-		       (andmap
-			(lambda (line)
-			  (and (or (and (list? line) (>= (length line) 2))
-				   (form-error "clause does not have at least 2 parts" line))
-			       (let ((name (car line)))
-				 (or (list? (cadr line))
-				     (make:line-error "second part of clause is not a sequence" (cadr line) name)))))
-			spec))
-		  `(make/proc (list ,@(map (lambda (line)
-					     `(list ,(car line)
-						    (list ,@(cadr line))
-						    ,@(let ((l (cddr line)))
-							(if (null? l)
-							    '()
-							    `((lambda ()
-								,@l))))))
-					   spec))
-			      ,(if (vector? argv) `',argv (car argv)))))))
-    (if (pair? argv)
-	(make spec argv)
-	(make spec '#())) ) )
+(define-macro (make spec #!optional (argv '()))
+  (let ((form-error (lambda (s . p) (apply error s spec p))))
+    (and (or (list? spec) (form-error "illegal specification (not a sequence)"))
+	 (or (pair? spec) (form-error "empty specification"))
+	 (andmap
+	  (lambda (line)
+	    (and (or (and (list? line) (>= (length line) 2))
+		     (form-error "clause does not have at least 2 parts" line))
+		 (let ((name (car line)))
+		   (or (list? (cadr line))
+		       (make:line-error "second part of clause is not a sequence" (cadr line) name)))))
+	  spec))
+    `(make/proc (list ,@(map (lambda (line)
+			       `(list ,(car line)
+				      (list ,@(cadr line))
+				      ,@(let ((l (cddr line)))
+					  (if (null? l)
+					      '()
+					      `((lambda ()
+						  ,@l))))))
+			     spec))
+		,argv)))
 
 (define (usage)
   (display #<<EOF
@@ -414,6 +414,11 @@ usage: chicken-setup [OPTION ...] FILENAME
   -t  -test EXTENSION ...        return success if all given extensions are installed
       -ls EXTENSION              list installed files for extension
       -fetch-tree                download and show repository catalog
+      -tree FILENAME             use repository catalog from given file
+      -svn URL                   fetch extension from subversion repository
+      -local PATH                fetch extension from local filesystem
+      -destdir PATH              specify alternative installation prefix
+      -revision REV              specify SVN revision for checkout
   --                             ignore all following arguments
 
   Builds and installs extension libraries.
@@ -422,6 +427,9 @@ EOF
   )
   (exit) )
 
+(define (make-setup-info-pathname fn #!optional (rpath (repository-path)))
+  (make-pathname rpath fn setup-file-extension) )
+
 (define program-path (make-parameter *install-bin-path*))
 
 (define installation-prefix
@@ -453,9 +461,12 @@ EOF
 		      '() ) 
 		,@info)) )
       (when (setup-verbose-flag) (printf "writing info ~A -> ~S ...~%" id info))
-      (let ((sid (->string id)))
-	(with-output-to-file (make-pathname (repo-path) sid setup-file-extension)
-	  (cut pp info) ) ) ) ) )
+      (let* ((sid (->string id))
+	    (setup-file (make-setup-info-pathname sid (repo-path #t)))
+	    (write-setup-info (with-output-to-file setup-file
+				(cut pp info))))
+	(unless *windows-shell* (run (chmod a+r ,setup-file)))
+	write-setup-info))))
 
 (define (fix-exports id info)
   (let-values (((einfo oinfo) (partition (lambda (item) (eq? 'exports (car item))) info)))
@@ -478,18 +489,8 @@ EOF
 	  (values `((exports ,@exports)) oinfo)
 	  (values '() oinfo) ) ) ) )
 
-(define (simple-install filename)
-  (let ((so (pathname-replace-extension filename ##sys#load-dynamic-extension))
-	(id (pathname-strip-extension filename)) )
-    (run (csc -O2 -no-trace -s ,filename -emit-exports ,(make-pathname #f id "exports")))
-    (when (setup-install-flag)
-      (unless *windows* (run (,*remove-command* ,(make-pathname (repo-path) so))))
-      (run (,*copy-command* ,so ,(repo-path)))
-      (write-info id (list (make-pathname (repository-path) so)) '()) )
-    (unless *keep-stuff* (run (rm ,so)) ) ) )
-
 (define (compute-tmpdir fname)
-  (if (string=? "egg-dir" (pathname-extension fname))
+  (if (equal? "egg-dir" (pathname-extension fname))
       fname
       (string-append fname "-dir") ) )
 
@@ -498,7 +499,8 @@ EOF
   (change-directory dir) )
 
 (define (rmtmpdir)
-  (chdir *base-directory*)
+  (unless (string=? (current-directory) *base-directory*)
+    (chdir *base-directory*) )
   (when *tmpdir-created*
     (set! *tmpdir-created* #f)
     (unless *keep-stuff*
@@ -512,7 +514,9 @@ EOF
     (with-input-from-file fn
       (lambda () (string=? "\x1f\x8b" (read-string 2))) ) )
   (let ((tmpdir (compute-tmpdir filename)))
-    (cond ((file-exists? tmpdir) (chdir tmpdir))
+    (cond ((file-exists? tmpdir) 
+	   (chdir tmpdir)
+	   (setup-build-directory (current-directory)) )
 	  (else
 	   (create-directory tmpdir)
 	   (set! *tmpdir-created* #t)
@@ -556,14 +560,21 @@ EOF
 	   (_ (error "invalid file-specification" f)) ) )
        flist) )
 
+(define (translate-extension f #!optional default)
+  (pathname-replace-extension f
+   (match (pathname-extension f)
+     (#f default)
+     ("so" ##sys#load-dynamic-extension)
+     ("o" (if *windows-shell* "obj" "o"))
+     ("a" (if *windows-shell* "lib" "a"))
+     (x x) ) ) )
+
 (define (install-extension id files #!optional (info '()))
-  (define (soify f)
-    (if (string=? (pathname-extension f) "so")
-	(pathname-replace-extension f ##sys#load-dynamic-extension)
-	f) )
+  (define (soify f) (translate-extension f))
   (when (setup-install-flag)
     (let* ((files (check-filelist (if (list? files) files (list files))))
 	   (rpath (repo-path))
+	   (rpathd (repo-path #t))
 	   (files (if *windows*
 		      (map (lambda (f) 
 			     (if (pair? f)
@@ -573,22 +584,37 @@ EOF
 		      files) ) 
 	   (dests (map (lambda (f)
 			 (let ((from (if (pair? f) (car f) f))
-			       (to (make-dest-pathname rpath f)) )
-			   (when (and (not *windows*) (string=? "so" (pathname-extension to)))
+			       (to (make-dest-pathname rpathd f)) )
+			   (when (and (not *windows*) (equal? "so" (pathname-extension to)))
 			     (run (,*remove-command* ,to)) )
-			   (copy-file from to) 
-			   to) )
+			   (copy-file from to)
+			   (unless *windows-shell*
+				   (run (chmod a+r ,to)))
+			   (and-let* ((static (assq 'static info)))
+			     (when (and (eq? (software-version) 'macosx)
+					(equal? (cadr static) from) )
+			       (run (ranlib ,to)) ) )
+			   (make-dest-pathname rpath f)))
 		       files) ) )
       (and-let* ((docs (assq 'documentation info)))
-	(print "\n* The following documentation files have been installed in " (doc-index) ":")
-	(for-each (cut print "  * " <>) (cdr docs)) 
-	(newline)
-	(set! *rebuild-doc-index* #t))
+	(let ((docpath (pathname-directory (doc-index #t))))
+	  (print "\n* Installing documentation files in " docpath ":")
+	  (for-each
+	   (lambda (f)
+	     (if (file-exists? f)
+		 (copy-file f (make-pathname docpath f))
+		 (print "Warning: file " f " doesn't seem to exist") )
+	     (print "  * " <>) (cdr docs)) 
+	   (newline)
+	   (set! *rebuild-doc-index* #t)) ) )
       (and-let* ((exs (assq 'examples info)))
-	(print "\n* The following example files have been installed in " *example-directory* ":")
+	(print "\n* Installing example files in " *example-directory* ":")
 	(for-each 
 	 (lambda (f)
 	   (copy-file f (make-pathname *example-directory* f))
+	   (unless *windows-shell*
+		   (run (chmod a+rx ,*example-directory*)))
+
 	   (print "  * " f) )
 	 (cdr exs))
 	(newline) )
@@ -596,12 +622,12 @@ EOF
 
 (define (install-program id files #!optional (info '()))
   (define (exify f)
-    (if (not (pathname-extension f)) 
-	(pathname-replace-extension f "exe")
-	f) )
+    (translate-extension
+     f
+     (if *windows-shell* "exe" #f) ) )
   (when (setup-install-flag)
     (let* ((files (check-filelist (if (list? files) files (list files))))
-	   (ppath (program-path)) 
+	   (ppath (if *destdir* (make-pathname *destdir* "bin") (program-path)))
 	   (files (if *windows*
                       (map (lambda (f)
                              (if (list? f) 
@@ -613,6 +639,8 @@ EOF
 			 (let ((from (if (pair? f) (car f) f))
 			       (to (make-dest-pathname ppath f)) )
 			   (copy-file from to) 
+			   (unless *windows-shell*
+				   (run (chmod a+r ,to)))
 			   to) )
 		       files) ) )
       (write-info id dests info) ) ) )
@@ -625,10 +653,12 @@ EOF
 			  (let ((from (if (pair? f) (car f) f))
 				(to (make-dest-pathname ppath f)) )
 			    (copy-file from to) 
+			    (unless *windows-shell*
+				    (run (chmod a+r ,to)))
 			    to) )
 			files) ) )
       (unless *windows-shell*
-	(run (chmod a+x ,(string-intersperse pfiles " "))) )
+	(run (chmod a+rx ,(string-intersperse pfiles " "))) )
       (write-info id pfiles info) ) ) )
 
 (define (uninstall-extension ext)
@@ -645,11 +675,14 @@ EOF
 	   (cdr files) ) )
 	(print "no files to uninstall") )
     (when (assq 'documentation info) (set! *rebuild-doc-index* #t))
-    (let ((sfile (make-pathname (repository-path) (->string ext) setup-file-extension)))
-      (delete-file* sfile) ) ) )
+    (delete-file* (make-setup-info-pathname (->string ext)))))
 
-(define (repo-path)
-  (let ((p (repository-path)))
+(define (repo-path #!optional ddir?)
+  (let ((p (if (and ddir? *destdir*)
+	       (make-pathname 
+		(list *destdir* "lib/chicken") 
+		(pathname-file (repository-path))) ; we need the binary-compat. version
+	       (repository-path))) )
     (ensure-directory p)
     p) )
 
@@ -658,9 +691,12 @@ EOF
     (if (file-exists? dir)
 	(unless (directory? dir)
 	  (error "can not create directory: a file with the same name already exists") )
-	(create-directory dir) ) ) )
+	(begin
+	  (create-directory dir)
+	  (unless *windows-shell*
+		  (run (chmod a+x ,dir)))))))
 
-(define (test-compile code #!key c++ (cc (if c++ *cxx* *cc*)) (cflags "") (ldflags "") (verb (setup-verbose-flag)) (compile-only #f))
+(define (try-compile code #!key c++ (cc (if c++ *cxx* *cc*)) (cflags "") (ldflags "") (verb (setup-verbose-flag)) (compile-only #f))
   (let* ((fname (create-temporary-file "c"))
 	 (oname (pathname-replace-extension fname "o"))
 	 (r (begin
@@ -683,6 +719,37 @@ EOF
     (system (sprintf "~A ~A" *remove-command* (quotewrap fname)))
     (zero? r) ) )
 
+(define (required-chicken-version v)
+  (when (string-ci<? (chicken-version) (->string v))
+    (error (sprintf "CHICKEN version ~a or higher is required" v)) ) )
+
+(define (upgrade-message ext msg)
+  (error
+   (sprintf
+    "the required extension `~s' ~a - please run~%~%  chicken-setup ~a~%~%and repeat the current installation operation."
+    ext msg ext) ) )
+
+(define (required-extension-version . args)
+  (let loop ((args args))
+    (match args
+      (() #f)
+      ((ext version . more)
+       (let ((info (extension-information ext))
+	     (version (->string version)) )
+	 (if info
+	     (let ((ver (and (assq 'version info) (cadr (assq 'version info)))))
+	       (cond ((not ver) (upgrade-message ext "has no associated version information"))
+		     ((string-ci<? (->string ver) version)
+		      (upgrade-message 
+		       ext
+		       (sprintf "is older than ~a, which is what this extension requires"
+				version) ) )
+		     (else (loop more)) ) ) 
+	     (upgrade-message ext "is not installed") ) ) )
+      (_ (error 'required-extension-information "bad argument format" args)) ) ) )
+
+(define test-compile try-compile)
+
 (define (find-library name proc)
   (test-compile 
    (sprintf "#ifdef __cplusplus~%extern \"C\"~%#endif~%char ~a();~%int main() { ~a(); return 0; }~%" proc proc)
@@ -715,8 +782,8 @@ EOF
 
 (define (download-repository-tree)
   (unless *repository-tree*
-    (print "downloading catalog ...")
-    (let loop ((hosts repository-hosts))
+    (when (setup-verbose-flag) (print "downloading catalog ..."))
+    (let loop ((hosts *repository-hosts*))
       (if (null? hosts)
 	  (error "unable to connect")
 	  (match hosts
@@ -725,7 +792,8 @@ EOF
 	      (lambda (return)
 		(or (handle-exceptions ex
 		      (begin (printf "could not connect to ~A.~%" host) #f)
-		      (printf "downloading catalog from ~A ...~%" host)
+		      (when (setup-verbose-flag)
+			(printf "downloading catalog from ~A ...~%" host) )
 		      (let-values (((i o) (setup-tcp-connect host port)))
 			(set! *last-decent-host* (car hosts))
 			(let ((req (http-get-request path remote-repository-name host)))
@@ -742,6 +810,7 @@ EOF
 			    (if (string=? "" ln)
 				(begin
 				  (set! *repository-tree* (read i))
+				  (set! *repository-tree-downloaded* #t)
 				  (when *debug*
 				    (print "catalog:")
 				    (pp *repository-tree*) )
@@ -749,7 +818,8 @@ EOF
 				  (close-output-port o)
 				  #t)
 				(loop) ) ) ) ) )
-		    (loop more) ) ) ) ) ) ) ) ) )
+		    (loop more) ) ) ) )
+	    ((x . _) (error "(internal) invalid host" x)) ) ) ) ) )
 
 (define *progress-indicator*
   (thread-start!
@@ -769,36 +839,72 @@ EOF
 	(thread-suspend! *progress-indicator*) ) ) )
 
 (define (download-data hostdata item #!optional filename)
-  (match hostdata
-    ((host path port)
-     (let ((fname (or filename (third (assq item *repository-tree*)))))
-       (printf "downloading ~A from ~A ~!" fname hostdata)
-       (let-values (((i o) (setup-tcp-connect host port)))
-	 (let ((req (http-get-request 
-		     (if filename (pathname-directory filename) path)
-		     (if filename (pathname-strip-directory fname) fname)
-		     host) ) )
-	   (when (setup-verbose-flag) (display req))
-	   (display req o) )
-	 (let loop ()
-	   (let ((ln (read-line i)))
-	     ;; check for 404 here...
-	     (if (string=? "" ln)
-		 (let ((data (with-progress-indicator (cut read-string #f i))))
-		   (close-input-port i)
-		   (close-output-port o)
-		   (with-output-to-file (pathname-strip-directory fname)
-		     (cut display data) binary:) ) 
-		 (loop) ) ) ) ) ) ) ) )
+  (unless hostdata (set! hostdata (car *repository-hosts*)))
+  (cond (*local-repository*
+	 (when (setup-verbose-flag) (printf "fetching from local directory ~a ...~%" *local-repository*))
+	 (let ((p (->string item)))
+	   (copy-file (make-pathname *local-repository* p) (make-pathname #f p "egg-dir")) ) )
+	(*svn-repository*
+	 (when (setup-verbose-flag) (printf "fetching from svn repository ~a ...~%" *svn-repository*))
+	 (let ((p (->string item)))
+	   (run (svn co ,(if *revision* (conc "--revision " *revision*) "")
+		     ,(make-pathname *svn-repository* p) ,(make-pathname #f p "egg-dir"))) ) )
+	(else
+	 (match hostdata
+	   ((host path port)
+	    (let ((fname (or filename (third (assq item *repository-tree*)))))
+	      (printf "downloading ~A from ~A ~!" fname hostdata)
+	      (let-values (((i o) (setup-tcp-connect host port)))
+		(let ((req (http-get-request 
+			    (if filename (pathname-directory filename) path)
+			    (if filename (pathname-strip-directory fname) fname)
+			    host) ) )
+		  (when *debug* (display req))
+		  (display req o) )
+		(let loop ()
+		  (let ((ln (read-line i)))
+		    ;; check for 404 here...
+		    (if (string=? "" ln)
+			(let ((data (with-progress-indicator (cut read-string #f i))))
+			  (close-input-port i)
+			  (close-output-port o)
+			  (with-output-to-file (pathname-strip-directory fname)
+			    (cut display data) 
+			    binary:) ) 
+			(loop) ) ) ) ) ) )
+	   (x (error "(internal) invalid host" x)) ) ) ) )
 
-(define (fetch-file-from-net ext)
+(define (fetch-file ext)
+  (define (eval-req r)
+    (when (setup-verbose-flag) 
+      (print "Testing system:")
+      (pp r) )
+    (let ((f (eval r)))
+      (when (setup-verbose-flag) 
+	(print "\t-> " f) )
+      f) )
   (define (requirements reqs)
     (fold 
      (lambda (r reqs)
-       (let ((node (assq r *repository-tree*)))
-	 (cond (node (append (requirements (cdddr node)) (list (car node)) reqs))
-	       ((memq r ##sys#core-library-modules) reqs)
-	       (else (error "Broken dependencies: extension does not exist" r) ) ) ) ) 
+       (cond ((symbol? r)
+	      (let ((node (assq r *repository-tree*)))
+		(cond (node (append (requirements (cdddr node)) (list (car node)) reqs))
+		      ((memq r ##sys#core-library-modules) reqs)
+		      (else (error "Broken dependencies: extension does not exist" r) ) ) ) )
+	     ((and *repository-tree-downloaded*
+		   (not *windows*)
+		   (or (zero? (current-user-id))
+		       (not *dont-ask*) ) )
+	      (print "WARNING: executing system test retrieved through potentially insecure network:\n")
+	      (pp r) 
+	      (cond ((yes-or-no?
+		      "Do you want to execute this code ?"
+		      (if (zero? (current-user-id)) "no" "yes") )
+		     (requirements (eval-req r) ) )
+		    (else
+		     (print "Test cancelled - aborting")
+		     (abort-setup) ) ) )
+	     (else (requirements (eval-req r) )) ) ) 
      '() 
      reqs) )
   (and (or *dont-ask*
@@ -807,7 +913,7 @@ EOF
 	    "yes") )
        (cond ((pathname-directory ext)
 	      (printf "Warning: no repository index available, trying direct download...~%" ext)
-	      (set! *last-decent-host* (car repository-hosts))
+	      (set! *last-decent-host* (car *repository-hosts*))
 	      (set! *dont-ask* #t)
 	      (download-data
 	       *last-decent-host*
@@ -833,17 +939,34 @@ EOF
   (let ((df (not *fetch-only*)))
     (let loop ((filename filename))
       (cond ((and df (with-ext filename "setup")) => run-setup-script)
-	    ((and df (with-ext filename "scm")) => simple-install)
-	    ((and df (or (with-ext filename "egg") (with-ext filename "egg-dir"))) =>
+	    ((or (with-ext filename "egg") (with-ext filename "egg-dir")) =>
 	     (lambda (f)
-	       (unpack/enter f)
-               (loop (pathname-replace-extension f "setup"))
-               (rmtmpdir) ) )
-	    ((fetch-file-from-net filename) 
+	       (when df
+		 (unpack/enter f)
+		 (let ((sfile (pathname-replace-extension f "setup")))
+		   (when (and (not (file-exists? sfile)) (file-exists? "tags") )
+		     (let ((ds (sort (directory "tags") string>=?)))
+		       (when (pair? ds) 
+			 (let ((d (make-pathname "tags" (car ds))))
+			   (chdir d)
+			   (setup-build-directory d) ) ) ) )
+		   (loop sfile)
+		   (rmtmpdir) ) ) ) )
+	    ((fetch-file filename) 
 	     (when df (loop (pathname-file filename))) ) ) ) ) )
 
-(define (doc-index)
-  (make-pathname (repo-path) "index.html"))
+(define (doc-index #!optional ddir?)
+  (make-pathname (repo-path ddir?) "index.html"))
+
+(define (extension-documented? rpath fn)
+  (let ([pn (make-setup-info-pathname fn rpath)])
+    (and (file-exists? pn)
+	 (with-input-from-file pn
+	   (lambda ()
+	     (not (not (alist-ref 'documentation (read) eq?))) ) ) ) ) )
+
+(define (delete-undocumented-extensions rpath lst)
+  (filter (cut extension-documented? rpath <>) lst) )
 
 (define (build-doc-index)
   (let ((rpath (repository-path))
@@ -874,9 +997,11 @@ EOF
 		 (and-let* ((r (assq 'release info)))
 		   (printf " Release: ~a" (cadr r)))
 		 (display "</td></tr>\n") ) ) )
-	   (sort (delete-duplicates
-		  (grep "^[^.].*\\.*$" (map pathname-file (directory rpath))) string=?)
-		 string<?) )
+	   (delete-undocumented-extensions 
+	    rpath
+	    (sort (delete-duplicates
+		   (grep "^[^.].*\\.*$" (map pathname-file (directory rpath))) string=?)
+		  string<?) ) )
 	  (display "</table></center></body></font></html>\n") ) ) ) ) )
 
 (define (format-string str cols #!optional right (padc #\space))
@@ -921,11 +1046,11 @@ EOF
 
 (define (main args)
   (define (parse-host host eggdir)
-    (set! repository-hosts
+    (set! *repository-hosts*
       (cons (match (string-match "(.+)\\:([0-9]+)" host)
 	      ((_ host port) (list host (if eggdir "eggs" "") (string->number port)))
 	      (_ (list host (if eggdir "eggs" "") 80)) )
-	    repository-hosts) )  )
+	    *repository-hosts*) )  )
   (setup-root-directory *base-directory*)
   (let ((uinst #f)
 	(anydone #f))
@@ -957,16 +1082,24 @@ EOF
 	(("-repository" dir . more)
 	 (repository-path dir)
 	 (loop more) )
+	(("-tree" file . more)
+	 (set! *repository-tree* (with-input-from-file file read))
+	 (loop more) )
 	(("--" . more)
 	 (exit) )
 	(("-program-path")
 	 (print (program-path))
 	 (exit) )
+	(("-destdir" path . more)
+	 (set! *example-directory* (make-pathname path "examples"))
+	 (set! *destdir* path) 
+	 (installation-prefix path)
+	 (loop more) )
 	(("-program-path" dir . more)
 	 (program-path dir)
 	 (loop more) )
 	(("-version" . _)
-	 (printf "chicken-setup - Version ~A~%" (chicken-version))
+	 (printf "chicken-setup - ~A~%" (chicken-version #t))
 	 (exit) )
 	(("-release" . _)
 	 (print (chicken-version))
@@ -984,8 +1117,7 @@ EOF
 	 (loop more) )
 	(("-host" host . more)
 	 (match (string-match "http://(.*)" host)
-	   ((_ host)
-	    (parse-host host #t) )
+	   ((_ host) (parse-host host #t) )
 	   (_ (parse-host host #t)) )
 	 (loop more) )
 	(("-proxy" proxy . more)
@@ -1024,10 +1156,21 @@ EOF
 	(("-debug" . more)
 	 (set! *debug* #t)
 	 (loop more) )
+	(("-revision" rev . more)
+	 (set! *revision* rev)
+	 (loop more) )
 	(("-check" . more)
 	 (set! *check-repository* #t)
 	 (set! anydone #t)
 	 (loop more) )
+	(("-svn" url . more)
+	 (set! *svn-repository* url)
+	 (set! *dont-ask* #t)
+	 (loop more) )
+	(("-local" path . more)
+	 (set! *local-repository* path)
+	 (set! *dont-ask* #t)
+	 (loop more) )
 	(("-fetch-tree" . more)
 	 (set! *fetch-tree-only* #t)
 	 (set! anydone #t)
@@ -1038,7 +1181,7 @@ EOF
 		  (print "the following extensions are currently not installed: " missing)
 		  1)
 		 (else (exit 0)) ) ) )
-	(((or "-run" "-script" "-proxy" "-host" "-csc-option"))
+	(((or "-run" "-script" "-proxy" "-host" "-csc-option" "-ls" "-destdir" "-tree" "-local" "-svn" "-eval"))
 	 (error "missing option argument" (car args)) )
 	((filename . more)
 	 (cond ((and (> (string-length filename) 0) (char=? #\- (string-ref filename 0)))
@@ -1080,6 +1223,7 @@ EOF
       (exit -1) )
   (call/cc
    (lambda (return)
-     (set! *abort-yes-or-no* return)
-     (main (command-line-arguments)) ) )
+     (set! *abort-hook* return)
+     (main (append (string-split (or (getenv "CHICKEN_SETUP_OPTIONS") ""))
+		   (command-line-arguments) ) ) ) )
   (rmtmpdir) )
