@@ -1,6 +1,6 @@
 ;;;; chicken.scm - The CHICKEN Scheme compiler (loader/main-module)
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -79,7 +79,7 @@
   (include "chicken-more-macros")
   (include "chicken-ffi-macros") )
 
-(##sys#provide 'extras 'srfi-1 'srfi-4)
+;;(##sys#provide 'extras 'srfi-1 'srfi-4)
 
 
 ;;; Prefix argument list with default options:
