@@ -127,6 +127,7 @@ The value returned is the value of the last form in BODY."
                            "define-optionals"
                            "define-reader-ctor"
                            "define-record"
+                           "defstruct"
                            "define-record-printer"
                            "define-record-type"
                            "define-record-scheme"
@@ -160,7 +161,8 @@ The value returned is the value of the last form in BODY."
               "call-with-values"
               "else"
               "foreign-lambda*" "foreign-safe-lambda*" "foreign-primitive"
-              "foreign-lambda" "foreign-safe-lambda" "foreign-code" "foreign-safe-wrapper"
+	      "foreign-declare"
+              "foreign-lambda" "foreign-safe-lambda" "foreign-code"
               "match" "match-lambda" "match-lambda*" "match-define" "match-let" "match-let*"
 
               "case" "case-lambda" "cond" "cond-expand" "condition-case" "switch"
@@ -175,7 +177,7 @@ The value returned is the value of the last form in BODY."
               "and-let*" "let-optionals" "let-optionals*" "let-macro"
               "fluid-let" "let-values" "let*-values" "letrec-values"
               "parameterize"
-              "module" "import-only" "import"
+              "module" "import-only" "import" "import*"
 
               "and" "or" "delay" "andmap" "ormap" "receive"
 
@@ -185,6 +187,8 @@ The value returned is the value of the last form in BODY."
               "with-input-from-pipe" "with-output-to-pipe"
               "with-input-from-string" "with-output-to-string"
 
+	      "loop"
+
               "declare" "include" "require-extension" "require" "require-for-syntax" "use" "quasiquote"
 
               "map" "syntax" "with-syntax" "syntax-case" "identifier-syntax" "syntax-rules") t)
