@@ -1,6 +1,6 @@
 ;;;; lolevel.scm - Low-level routines for CHICKEN
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -46,13 +46,6 @@
 #ifndef C_NONUNIX
 # include <sys/mman.h>
 #endif
-#if defined(__i386__) && !defined(C_NONUNIX) && !defined(__CYGWIN__)
-# define C_valloc(n)               valloc(n)
-#elif defined(__i386__) || defined(_M_IX86)
-# define C_valloc(n)               malloc(n)
-#else
-# define C_valloc(n)               NULL
-#endif
 
 #define C_pointer_to_object(ptr)   ((C_word*)C_block_item(ptr, 0))
 #define C_w2b(x)                   C_fix(C_wordstobytes(C_unfix(x)))
@@ -132,7 +125,8 @@ EOF
 		     [else (xerr to)] ) ]
 	      [(or (##sys#bytevector? from) (string? from))
 	       (let ([nfrom (##sys#size from)])
-		 (cond [(##sys#pointer? to) (memmove2 to from (checkn (:optional n nfrom) nfrom))]
+		 (cond [(or (##sys#pointer? to) (##sys#locative? to))
+			(memmove2 to from (checkn (:optional n nfrom) nfrom))]
 		       [(or (##sys#bytevector? to) (string? to))
 			(memmove4 to from (checkn2 (:optional n nfrom) nfrom (##sys#size to))) ]
 		       [else (xerr to)] ) ) ]
@@ -411,7 +405,7 @@ EOF
       [malloc
        (foreign-lambda* scheme-object ((int size))
 	 "char *bv;
-           if((bv = (char *)malloc(size + 3 + sizeof(C_header))) == NULL) return(C_SCHEME_FALSE);
+           if((bv = (char *)C_malloc(size + 3 + sizeof(C_header))) == NULL) return(C_SCHEME_FALSE);
            bv = (char *)C_align((C_word)bv);
            ((C_SCHEME_BLOCK *)bv)->header = C_BYTEVECTOR_TYPE | size;
            return((C_word)bv);") ] )
