@@ -2,18 +2,18 @@
 ;
 ; By Sergey Khorev
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
 ; conditions are met:
 ;
 ;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
-;     disclaimer. 
+;     disclaimer.
 ;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
-;     disclaimer in the documentation and/or other materials provided with the distribution. 
+;     disclaimer in the documentation and/or other materials provided with the distribution.
 ;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
-;     products derived from this software without specific prior written permission. 
+;     products derived from this software without specific prior written permission.
 ;
 ; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 ; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
@@ -25,7 +25,7 @@
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
 ;
-; Send bugs, suggestions and ideas to: 
+; Send bugs, suggestions and ideas to:
 ;
 ; felix@call-with-current-continuation.org
 ;
@@ -41,8 +41,7 @@
 ; perm/isvtx  perm/isuid  perm/isgid
 ; file-select
 ; symbolic-link?
-; signal/...
-; set-signal-handler!  set-signal-mask!
+; set-signal-mask!  signal-mask  signal-masked?  signal-mask!  signal-unmask!
 ; user-information  group-information  get-groups  set-groups!  initialize-groups
 ; errno/wouldblock
 ; change-file-owner
@@ -51,7 +50,7 @@
 ; create-session
 ; process-group-id  set-process-group-id!
 ; create-symbolic-link  read-symbolic-link
-; file-truncate 
+; file-truncate
 ; file-lock  file-lock/blocking  file-unlock  file-test-lock
 ; create-fifo  fifo?
 ; prot/...
@@ -60,7 +59,7 @@
 ; set-alarm!
 ; terminal-port?  terminal-name
 ; process-fork  process-wait
-; parent-process-id 
+; parent-process-id
 ; process-signal
 
 
@@ -69,7 +68,7 @@
   (uses scheduler regex extras utils)
   (disable-interrupts)
   (usual-integrations)
-  (hide ##sys#stat yield close-handle)
+  (hide ##sys#stat close-handle posix-error)
   (foreign-declare #<<EOF
 #ifndef WIN32_LEAN_AND_MEAN
 # define WIN32_LEAN_AND_MEAN
@@ -88,6 +87,7 @@ However, the _MSC_VER test is still needed for vcbuild.bat.
 #include <winsock.h>
 #endif
 
+#include <signal.h>
 #include <errno.h>
 #include <io.h>
 #include <stdio.h>
@@ -112,12 +112,14 @@ static C_TLS int C_pipefds[ 2 ];
 static C_TLS time_t C_secs;
 static C_TLS struct tm C_tm;
 static C_TLS struct stat C_statbuf;
+
 /* pipe handles */
 static C_TLS HANDLE C_rd0, C_wr0, C_wr0_, C_rd1, C_wr1, C_rd1_;
 static C_TLS HANDLE C_save0, C_save1; /* saved I/O handles */
 static C_TLS char C_rdbuf; /* one-char buffer for read */
 static C_TLS int C_exstatus;
 
+/* platform information */
 static C_TLS char C_hostname[256];
 static C_TLS char C_osver[16];
 static C_TLS char C_osrel[16];
@@ -179,6 +181,7 @@ static DIR *opendir(const char *name)
     free(what);
     return dir;
 }
+
 static int closedir(DIR * dir)
 {
     if (dir)
@@ -189,6 +192,7 @@ static int closedir(DIR * dir)
     }
     return -1;
 }
+
 static struct dirent *readdir(DIR * dir)
 {
     if (dir)
@@ -246,7 +250,7 @@ static C_word C_fcall C_setenv(C_word x, C_word y);
 C_word C_fcall C_setenv(C_word x, C_word y) {
   char *sx = C_data_pointer(x),
        *sy = C_data_pointer(y);
-  int n1 = C_strlen(sx), n2 = C_strlen(sy);				       
+  int n1 = C_strlen(sx), n2 = C_strlen(sy);
   char *buf = (char *)C_malloc(n1 + n2 + 2);
   if(buf == NULL) return(C_fix(0));
   else {
@@ -275,10 +279,10 @@ void C_fcall C_free_exec_args() {
   while((*a) != NULL) C_free(*(a++));
 }
 
-#define C_execvp(f)         C_fix(execvp(C_data_pointer(f), C_exec_args))
+#define C_execvp(f)         C_fix(execvp(C_data_pointer(f), (const char *const *)C_exec_args))
 
 /* MS replacement for the fork-exec pair */
-#define C_spawnvp(m, f)	    C_fix(spawnvp(C_unfix(m), C_data_pointer(f), C_exec_args))
+#define C_spawnvp(m, f)	    C_fix(spawnvp(C_unfix(m), C_data_pointer(f), (const char *const *)C_exec_args))
 
 #define C_open(fn, fl, m)   C_fix(open(C_c_string(fn), C_unfix(fl), C_unfix(m)))
 #define C_read(fd, b, n)    C_fix(read(C_unfix(fd), C_data_pointer(b), C_unfix(n)))
@@ -294,12 +298,16 @@ void C_fcall C_free_exec_args() {
 #define C_asctime(v)        (memset(&C_tm, 0, sizeof(struct tm)), C_tm.tm_sec = C_unfix(C_block_item(v, 0)), C_tm.tm_min = C_unfix(C_block_item(v, 1)), C_tm.tm_hour = C_unfix(C_block_item(v, 2)), C_tm.tm_mday = C_unfix(C_block_item(v, 3)), C_tm.tm_mon = C_unfix(C_block_item(v, 4)), C_tm.tm_year = C_unfix(C_block_item(v, 5)), C_tm.tm_wday = C_unfix(C_block_item(v, 6)), C_tm.tm_yday = C_unfix(C_block_item(v, 7)), C_tm.tm_isdst = (C_block_item(v, 8) != C_SCHEME_FALSE), asctime(&C_tm) )
 #define C_mktime(v)        (memset(&C_tm, 0, sizeof(struct tm)), C_tm.tm_sec = C_unfix(C_block_item(v, 0)), C_tm.tm_min = C_unfix(C_block_item(v, 1)), C_tm.tm_hour = C_unfix(C_block_item(v, 2)), C_tm.tm_mday = C_unfix(C_block_item(v, 3)), C_tm.tm_mon = C_unfix(C_block_item(v, 4)), C_tm.tm_year = C_unfix(C_block_item(v, 5)), C_tm.tm_wday = C_unfix(C_block_item(v, 6)), C_tm.tm_yday = C_unfix(C_block_item(v, 7)), C_tm.tm_isdst = (C_block_item(v, 8) != C_SCHEME_FALSE), (C_temporary_flonum = mktime(&C_tm)) != -1)
 
-/* mapping from Win32 error codes to errno */
+/*
+  mapping from Win32 error codes to errno
+*/
+
 typedef struct
 {
     DWORD   win32;
     int	    libc;
 } errmap_t;
+
 static errmap_t errmap[] =
 {
     {ERROR_INVALID_FUNCTION,      EINVAL},
@@ -371,6 +379,7 @@ static int zero_handles()
     C_save0 = C_save1 = INVALID_HANDLE_VALUE;
     return 1;
 }
+
 static int close_handles()
 {
     if (C_rd0 != INVALID_HANDLE_VALUE)
@@ -397,6 +406,7 @@ static int close_handles()
     }
     return zero_handles();
 }
+
 static int redir_io()
 {
     SECURITY_ATTRIBUTES sa;
@@ -428,6 +438,7 @@ static int redir_io()
     C_rd1 = INVALID_HANDLE_VALUE;
     return 1;
 }
+
 static int run_process(char *cmdline)
 {
     PROCESS_INFORMATION pi;
@@ -459,6 +470,7 @@ static int run_process(char *cmdline)
 	return 0;
     }
 }
+
 static int pipe_write(int hpipe, void* buf, int count)
 {
     DWORD done = 0;
@@ -470,6 +482,7 @@ static int pipe_write(int hpipe, void* buf, int count)
 	return 0;
     }
 }
+
 static int pipe_read(int hpipe)
 {
     DWORD done = 0;
@@ -487,12 +500,13 @@ static int pipe_read(int hpipe)
     set_errno(GetLastError());
     return 0;
 }
+
 static int pipe_ready(int hpipe)
 {
     DWORD avail = 0;
     if (PeekNamedPipe((HANDLE)hpipe, NULL, 0, NULL, &avail, NULL) && avail)
 	return 1;
-    else 
+    else
     {
 	Sleep(0); /* give pipe a chance */
 	if (PeekNamedPipe((HANDLE)hpipe, NULL, 0, NULL, &avail, NULL))
@@ -512,7 +526,7 @@ static int pipe_ready(int hpipe)
 #define close_handle(h) CloseHandle((HANDLE)h)
 
 int process_wait(int h, int t)
-{   
+{
     if (WaitForSingleObject((HANDLE)h, (t ? 0 : INFINITE)) == WAIT_OBJECT_0)
     {
 	DWORD ret;
@@ -550,7 +564,7 @@ int sysinfo()
     if (get_hostname() && GetVersionEx(&ovf))
     {
 	SYSTEM_INFO si;
-	_snprintf(C_osver, sizeof(C_osver) - 1, "%d.%d.%d", 
+	_snprintf(C_osver, sizeof(C_osver) - 1, "%d.%d.%d",
 			   ovf.dwMajorVersion, ovf.dwMinorVersion, ovf.dwBuildNumber);
 	switch (ovf.dwPlatformId)
 	{
@@ -569,8 +583,23 @@ int sysinfo()
 	switch (si.wProcessorArchitecture)
 	{
     	case PROCESSOR_ARCHITECTURE_INTEL:
-	    strncpy(C_processor, "Intel", sizeof(C_processor) - 1);
+	    strncpy(C_processor, "x86", sizeof(C_processor) - 1);
 	    break;
+#       ifdef PROCESSOR_ARCHITECTURE_IA64
+    	case PROCESSOR_ARCHITECTURE_IA64:
+	    strncpy(C_processor, "IA64", sizeof(C_processor) - 1);
+	    break;
+#       endif
+#       ifdef PROCESSOR_ARCHITECTURE_AMD64
+    	case PROCESSOR_ARCHITECTURE_AMD64:
+	    strncpy(C_processor, "x64", sizeof(C_processor) - 1);
+	    break;
+#       endif
+#       ifdef PROCESSOR_ARCHITECTURE_IA32_ON_WIN64
+    	case PROCESSOR_ARCHITECTURE_IA32_ON_WIN64:
+	    strncpy(C_processor, "WOW64", sizeof(C_processor) - 1);
+	    break;
+#       endif
     	case PROCESSOR_ARCHITECTURE_UNKNOWN:
 	default:
 	    strncpy(C_processor, "Unknown", sizeof(C_processor) - 1);
@@ -585,6 +614,150 @@ int sysinfo()
 #define C_get_hostname() (get_hostname() ? C_SCHEME_TRUE : C_SCHEME_FALSE)
 #define C_sysinfo() (sysinfo() ? C_SCHEME_TRUE : C_SCHEME_FALSE)
 
+/*
+    Spawn a process, either through shell or directly.
+    Params:
+    cmd         Command to execute.
+    env         Environment for the new process (may be NULL).
+    handle, stdin, stdout, stderr
+                Spawned process info are returned in integers.
+                When spawned process shares standard io stream with the parent
+                process the respective value in handle, stdin, stdout, stderr
+                is -1.
+    params      A bitmask controling operation.
+                Bit 1: Child & parent share standard input if this bit is set.
+                Bit 2: Share standard output if bit is set.
+                Bit 3: Share standard error if bit is set.
+                (Bit 4: Execute command in shell if bit is set.)
+
+    Returns: nonzero return value indicates failure.
+*/
+static
+int C_process(const char * cmd, const char ** env,
+              int * phandle, int * pstdin_fd, int * pstdout_fd, int * pstderr_fd,
+              int params)
+{
+    int exit_code = 0, i = 0;
+    const int
+        f_share_io[3] = { params & 1, params & 2, params & 4};
+#if 0
+    const int
+        f_use_shell = params & 8;
+#endif
+
+    char * buf = NULL;
+    const char * invoke_cmd = NULL;
+    int io_fds[3]={-1,-1,-1};
+    HANDLE
+        child_io_handles[3]={NULL,NULL,NULL},
+        standard_io_handles[3]={
+            GetStdHandle(STD_INPUT_HANDLE),
+            GetStdHandle(STD_OUTPUT_HANDLE),
+            GetStdHandle(STD_ERROR_HANDLE)};
+
+    const char modes[3]="rww";
+    HANDLE cur_process = GetCurrentProcess(), child_process = NULL;
+
+    /****** create io handles & fds ***/
+
+    for (i=0; i<3 && exit_code == 0; ++i)
+    {
+        if (f_share_io[i])
+        {
+            exit_code = !DuplicateHandle(
+                cur_process, standard_io_handles[i],
+                cur_process, &child_io_handles[i],
+                0, FALSE, DUPLICATE_SAME_ACCESS);
+        }
+        else
+        {
+            HANDLE a, b, parent_end;
+            exit_code = !CreatePipe(&a,&b,NULL,0);
+            if(0==exit_code)
+            {
+                if (modes[i]=='r') { child_io_handles[i]=a; parent_end=b; }
+                else { parent_end=a; child_io_handles[i]=b; }
+            }
+            exit_code=(io_fds[i]=_open_osfhandle((long)parent_end,0))<0;
+        }
+    }
+
+    /****** make handles inheritable */
+
+    for (i=0; i<3 && exit_code == 0; ++i)
+        exit_code = !SetHandleInformation(child_io_handles[i], HANDLE_FLAG_INHERIT, -1);
+
+    /****** create command line ******/
+
+#if 0
+    if (f_use_shell && exit_code == 0)
+    {
+        const char * shell = NULL;
+        static const char * const fmt = "%s /c %s";
+
+        shell=getenv("COMSPEC");
+        if (NULL==shell)
+        {
+            OSVERSIONINFO ovf;
+            ovf.dwOSVersionInfoSize = sizeof(ovf);
+            if (GetVersionEx(&ovf) && (ovf.dwPlatformId == VER_PLATFORM_WIN32_NT))
+                shell="cmd.exe";
+            else
+                shell="command.com";
+        }
+
+        buf = (char*) malloc(strlen(fmt)+strlen(shell)+strlen(cmd));
+        exit_code=(NULL==buf);
+        if (0==exit_code) { sprintf(buf,fmt,shell,cmd); invoke_cmd = buf; }
+    }
+    else
+#endif
+        invoke_cmd = cmd;
+
+    /****** finally spawn process ****/
+
+    if (0==exit_code)
+    {
+        PROCESS_INFORMATION pi;
+        STARTUPINFO si;
+
+        ZeroMemory(&pi,sizeof pi);
+        ZeroMemory(&si,sizeof si);
+        si.cb = sizeof si;
+        si.dwFlags = STARTF_USESTDHANDLES;
+        si.hStdInput = child_io_handles[0];
+        si.hStdOutput = child_io_handles[1];
+        si.hStdError = child_io_handles[2];
+
+        exit_code = !CreateProcess(
+            NULL,(char*)invoke_cmd,NULL,NULL,TRUE,0,(char**)env,NULL,&si,&pi);
+
+        if (0==exit_code)
+        {
+            child_process=pi.hProcess;
+            CloseHandle(pi.hThread);
+        }
+    }
+
+    /****** cleanup & return *********/
+
+    free(buf);
+    for (i=0; i<3; ++i) CloseHandle(child_io_handles[i]);
+    if (exit_code != 0)
+    {
+        for (i=0; i<3; ++i) _close(io_fds[i]);
+        set_errno(GetLastError());
+    }
+    else
+    {
+        *phandle = (int)child_process;
+        *pstdin_fd = io_fds[0];
+        *pstdout_fd = io_fds[1];
+        *pstderr_fd = io_fds[2];
+    }
+
+    return exit_code;
+}
 EOF
 ) )
 
@@ -595,8 +768,8 @@ EOF
     (no-bound-checks)
     (no-procedure-checks-for-usual-bindings)
     (bound-to-procedure
-     ##sys#make-port ##sys#file-info ##sys#update-errno ##sys#fudge ##sys#make-c-string ##sys#check-port 
-     ##sys#error ##sys#signal-hook ##sys#peek-unsigned-integer
+     ##sys#make-port ##sys#file-info ##sys#update-errno ##sys#fudge ##sys#make-c-string ##sys#check-port
+     ##sys#error ##sys#signal-hook ##sys#peek-unsigned-integer ##sys#process
      ##sys#peek-fixnum ##sys#make-structure ##sys#check-structure ##sys#enable-interrupts) ) ] )
 
 (cond-expand
@@ -614,18 +787,18 @@ EOF
     (define-macro (##sys#check-number . _) '(##core#undefined))
     (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
  [else
-  (declare (emit-exports "posixwin.exports"))] )
+  (declare (emit-exports "posix.exports"))] )
 
 (register-feature! 'posix)
 
-;;; We need this for threading:
+(define posix-error
+  (let ([strerror (foreign-lambda c-string "strerror" int)]
+        [string-append string-append] )
+    (lambda (type loc msg . args)
+      (let ([rn (##sys#update-errno)])
+        (apply ##sys#signal-hook type loc (string-append msg " - " (strerror rn)) args) ) ) ) )
 
-(define (yield)
-  (##sys#call-with-current-continuation
-   (lambda (return)
-     (let ((ct ##sys#current-thread))
-       (##sys#setslot ct 1 (lambda () (return (##core#undefined))))
-       (##sys#schedule) ) ) ) )
+(define ##sys#posix-error posix-error)
 
 
 ;;; Lo-level I/O:
@@ -692,7 +865,7 @@ EOF
 	(let ([fd (##core#inline "C_open" (##sys#make-c-string (##sys#expand-home-path filename)) flags mode)])
 	  (when (eq? -1 fd)
 	    (##sys#update-errno)
-	    (##sys#signal-hook #:file-error 'file-open "can not open file" filename flags mode) )
+	    (##sys#signal-hook #:file-error 'file-open "cannot open file" filename flags mode) )
 	  fd) ) ) ) )
 
 (define file-close
@@ -700,7 +873,7 @@ EOF
     (##sys#check-exact fd 'file-close)
     (when (fx< (##core#inline "C_close" fd) 0)
       (##sys#update-errno)
-      (##sys#signal-hook #:file-error 'file-close "can not close file" fd) ) ) )
+      (##sys#signal-hook #:file-error 'file-close "cannot close file" fd) ) ) )
 
 (define file-read
   (let ([make-string make-string] )
@@ -713,7 +886,7 @@ EOF
 	(let ([n (##core#inline "C_read" fd buf size)])
 	  (when (eq? -1 n)
 	    (##sys#update-errno)
-	    (##sys#signal-hook #:file-error 'file-read "can not read from file" fd size) )
+	    (##sys#signal-hook #:file-error 'file-read "cannot read from file" fd size) )
 	  (list buf n) ) ) ) ) )
 
 (define file-write
@@ -726,7 +899,7 @@ EOF
       (let ([n (##core#inline "C_write" fd buffer size)])
 	(when (eq? -1 n)
 	  (##sys#update-errno)
-	  (##sys#signal-hook #:file-error 'file-write "can not write to file" fd size) )
+	  (##sys#signal-hook #:file-error 'file-write "cannot write to file" fd size) )
 	n) ) ) )
 
 (define file-mkstemp
@@ -738,7 +911,7 @@ EOF
              [path-length (string-length buf)])
         (when (eq? -1 fd)
           (##sys#update-errno)
-          (##sys#signal-hook #:file-error 'file-mkstemp "can not create temporary file" template) )
+          (##sys#signal-hook #:file-error 'file-mkstemp "cannot create temporary file" template) )
         (values fd (##sys#substring buf 0 (fx- path-length 1) ) ) ) ) ) )
 
 
@@ -768,13 +941,14 @@ EOF
 		 [else (##sys#signal-hook #:type-error "bad argument type - not a fixnum or string" file)] ) ] )
     (when (fx< r 0)
       (##sys#update-errno)
-      (##sys#signal-hook #:file-error "can not access file" file) ) ) )
+      (##sys#signal-hook #:file-error "cannot access file" file) ) ) )
 
 (define (file-stat f #!optional link)
   (##sys#stat f)
   (vector _stat_st_ino _stat_st_mode _stat_st_nlink
 	  _stat_st_uid _stat_st_gid _stat_st_size
-	  _stat_st_atime _stat_st_ctime _stat_st_mtime) )
+	  _stat_st_atime _stat_st_ctime _stat_st_mtime
+	  0 0 0 0) )
 
 (define (file-size f) (##sys#stat f) _stat_st_size)
 (define (file-modification-time f) (##sys#stat f) _stat_st_mtime)
@@ -794,7 +968,7 @@ EOF
 
 (define file-position
   (lambda (port)
-    (let ([pos (cond [(port? port) 
+    (let ([pos (cond [(port? port)
 		      (if (eq? (##sys#slot port 7) 'stream)
 			  (##core#inline "C_ftell" port)
 			  -1) ]
@@ -802,7 +976,7 @@ EOF
 		     [else (##sys#signal-hook #:type-error 'file-position "invalid file" port)] ) ] )
       (when (fx< pos 0)
 	(##sys#update-errno)
-	(##sys#signal-hook #:file-error 'file-position "can not retrieve file position of port" port) )
+	(##sys#signal-hook #:file-error 'file-position "cannot retrieve file position of port" port) )
       pos) ) )
 
 (define set-file-position!
@@ -811,38 +985,38 @@ EOF
       (##sys#check-exact pos 'set-file-position!)
       (##sys#check-exact whence 'set-file-position!)
       (when (fx< pos 0) (##sys#signal-hook #:bounds-error 'set-file-position! "invalid negative port position" pos port))
-      (unless (cond [(port? port) 
+      (unless (cond [(port? port)
 		     (and (eq? (##sys#slot port 7) 'stream)
 			  (##core#inline "C_fseek" port pos whence) ) ]
 		    [(fixnum? port) (##core#inline "C_lseek" port pos whence)]
 		    [else (##sys#signal-hook #:type-error 'set-file-position! "invalid file" port)] )
 	(##sys#update-errno)
-	(##sys#signal-hook #:file-error 'set-file-position! "can not set file position" port pos) ) ) ) )
+	(##sys#signal-hook #:file-error 'set-file-position! "cannot set file position" port pos) ) ) ) )
 
 
 ;;; Directory stuff:
 
-(define create-directory 
+(define create-directory
   (lambda (name)
     (##sys#check-string name 'create-directory)
     (unless (zero? (##core#inline "C_mkdir" (##sys#make-c-string (##sys#expand-home-path name))))
       (##sys#update-errno)
-      (##sys#signal-hook #:file-error 'create-directory "can not create directory" name) ) ) )
+      (##sys#signal-hook #:file-error 'create-directory "cannot create directory" name) ) ) )
 
 (define change-directory
   (lambda (name)
     (##sys#check-string name 'change-directory)
     (unless (zero? (##core#inline "C_chdir" (##sys#make-c-string (##sys#expand-home-path name))))
       (##sys#update-errno)
-      (##sys#signal-hook #:file-error 'change-directory "can not change current directory" name) ) ) )
+      (##sys#signal-hook #:file-error 'change-directory "cannot change current directory" name) ) ) )
 
 (define delete-directory
   (lambda (name)
     (##sys#check-string name 'delete-directory)
     (unless (zero? (##core#inline "C_rmdir" (##sys#make-c-string (##sys#expand-home-path name))))
       (##sys#update-errno)
-      (##sys#signal-hook #:file-error 'delete-directory "can not delete directory" name) ) ) )
-  
+      (##sys#signal-hook #:file-error 'delete-directory "cannot delete directory" name) ) ) )
+
 (define directory
   (let ([string-append string-append]
 	[make-string make-string]
@@ -856,19 +1030,19 @@ EOF
 	(if (##sys#null-pointer? handle)
 	    (begin
 	      (##sys#update-errno)
-	      (##sys#signal-hook #:file-error 'directory "can not open directory" spec) )
+	      (##sys#signal-hook #:file-error 'directory "cannot open directory" spec) )
 	    (let loop ()
 	      (##core#inline "C_readdir" handle entry)
 	      (if (##sys#null-pointer? entry)
 		  (begin
 		    (##core#inline "C_closedir" handle)
-		    '() ) 
+		    '() )
 		  (let* ([flen (##core#inline "C_foundfile" entry buffer)]
 			 [file (##sys#substring buffer 0 flen)]
 			 [char1 (string-ref file 0)]
 			 [char2 (and (> flen 1) (string-ref file 1))] )
-		    (if (and (eq? char1 #\.) 
-			     (or (not char2) 
+		    (if (and (eq? char1 #\.)
+			     (or (not char2)
 				 (and (eq? char2 #\.) (eq? flen 2))
 				 (not show-dotfiles?) ) )
 			(loop)
@@ -889,7 +1063,7 @@ EOF
 	    (##sys#update-errno)
 	    (if len
 		(##sys#substring buffer 0 len)
-		(##sys#signal-hook #:file-error 'current-directory "can not retrieve current directory") ) ) ) ) ) )
+		(##sys#signal-hook #:file-error 'current-directory "cannot retrieve current directory") ) ) ) ) ) )
 
 
 ;;; Pipes:
@@ -900,7 +1074,7 @@ EOF
   (define (check cmd inp r)
     (##sys#update-errno)
     (if (##sys#null-pointer? r)
-	(##sys#signal-hook #:file-error "can not open pipe" cmd) 
+	(##sys#signal-hook #:file-error "cannot open pipe" cmd)
 	(let ([port (##sys#make-port inp ##sys#stream-port-class "(pipe)" 'stream)])
 	  (##core#inline "C_set_file_ptr" port r)
 	  port) ) )
@@ -923,7 +1097,7 @@ EOF
 	 (case m
 	   ((###text) (##core#inline_allocate ("open_text_output_pipe" 2) (##sys#make-c-string cmd)))
 	   ((###binary) (##core#inline_allocate ("open_binary_output_pipe" 2) (##sys#make-c-string cmd)))
-	   (else (badmode m)) ) ) ) ) ) 
+	   (else (badmode m)) ) ) ) ) )
   (set! close-input-pipe
     (lambda (port)
       (##sys#check-port port 'close-input-pipe)
@@ -983,9 +1157,72 @@ EOF
     (lambda ()
       (when (fx< (##core#inline "C_pipe" #f) 0)
 	(##sys#update-errno)
-	(##sys#signal-hook #:file-error 'create-pipe "can not create pipe") )
+	(##sys#signal-hook #:file-error 'create-pipe "cannot create pipe") )
       (values _pipefd0 _pipefd1) ) )
 
+;;; Signal processing:
+
+(define-foreign-variable _nsig int "NSIG")
+(define-foreign-variable _sigterm int "SIGTERM")
+(define-foreign-variable _sigint int "SIGINT")
+(define-foreign-variable _sigfpe int "SIGFPE")
+(define-foreign-variable _sigill int "SIGILL")
+(define-foreign-variable _sigsegv int "SIGSEGV")
+(define-foreign-variable _sigabrt int "SIGABRT")
+(define-foreign-variable _sigbreak int "SIGBREAK")
+
+(define signal/term _sigterm)
+(define signal/int _sigint)
+(define signal/fpe _sigfpe)
+(define signal/ill _sigill)
+(define signal/segv _sigsegv)
+(define signal/abrt _sigabrt)
+(define signal/break _sigbreak)
+(define signal/alrm 0)
+(define signal/chld 0)
+(define signal/cont 0)
+(define signal/hup 0)
+(define signal/io 0)
+(define signal/kill 0)
+(define signal/pipe 0)
+(define signal/prof 0)
+(define signal/quit 0)
+(define signal/stop 0)
+(define signal/trap 0)
+(define signal/tstp 0)
+(define signal/urg 0)
+(define signal/usr1 0)
+(define signal/usr2 0)
+(define signal/vtalrm 0)
+(define signal/winch 0)
+(define signal/xcpu 0)
+(define signal/xfsz 0)
+
+(define signals-list
+  (list
+    signal/term signal/int signal/fpe signal/ill
+    signal/segv signal/abrt signal/break))
+
+(let ([oldhook ##sys#interrupt-hook]
+      [sigvector (make-vector 256 #f)] )
+  (set! signal-handler
+    (lambda (sig)
+      (##sys#check-exact sig 'signal-handler)
+      (##sys#slot sigvector sig) ) )
+  (set! set-signal-handler!
+    (lambda (sig proc)
+      (##sys#check-exact sig 'set-signal-handler!)
+      (##core#inline "C_establish_signal_handler" sig (and proc sig))
+      (vector-set! sigvector sig proc) ) )
+  (set! ##sys#interrupt-hook
+    (lambda (reason state)
+      (let ([h (##sys#slot sigvector reason)])
+        (if h
+            (begin
+              (h reason)
+              (##sys#context-switch state) )
+            (oldhook reason state) ) ) ) ) )
+
 ;;; More errno codes:
 
 (define-foreign-variable _errno int "errno")
@@ -1076,7 +1313,7 @@ EOF
     (##sys#check-exact m 'change-file-mode)
     (when (fx< (##core#inline "C_chmod" (##sys#make-c-string (##sys#expand-home-path fname)) m) 0)
       (##sys#update-errno)
-      (##sys#signal-hook #:file-error 'change-file-mode "can not change file mode" fname m) ) ) )
+      (##sys#signal-hook #:file-error 'change-file-mode "cannot change file mode" fname m) ) ) )
 
 (define-foreign-variable _r_ok int "2")
 (define-foreign-variable _w_ok int "4")
@@ -1117,7 +1354,7 @@ EOF
   (define (check fd inp r)
     (##sys#update-errno)
     (if (##sys#null-pointer? r)
-	(##sys#signal-hook #:file-error "can not open file" fd) 
+	(##sys#signal-hook #:file-error "cannot open file" fd)
 	(let ([port (##sys#make-port inp ##sys#stream-port-class "(fdport)" 'stream)])
 	  (##core#inline "C_set_file_ptr" port r)
 	  port) ) )
@@ -1137,7 +1374,7 @@ EOF
 	(let ([fd (##core#inline "C_C_fileno" port)])
 	  (when (fx< fd 0)
 	    (##sys#update-errno)
-	    (##sys#signal-hook #:file-error 'port->fileno "can not access file-descriptor of port" port) )
+	    (##sys#signal-hook #:file-error 'port->fileno "cannot access file-descriptor of port" port) )
 	  fd)
 	(##sys#signal-hook #:type-error 'port->fileno "port has no attached file" port) ) ) )
 
@@ -1151,12 +1388,12 @@ EOF
 		    (##core#inline "C_dup2" old n) ) ) ] )
       (when (fx< fd 0)
 	(##sys#update-errno)
-	(##sys#signal-hook #:file-error 'duplicate-fileno "can not duplicate file descriptor" old) )
+	(##sys#signal-hook #:file-error 'duplicate-fileno "cannot duplicate file descriptor" old) )
       fd) ) )
 
 ;;; Environment access:
 
-(define setenv 
+(define setenv
   (lambda (var val)
     (##sys#check-string var 'setenv)
     (##sys#check-string val 'setenv)
@@ -1191,11 +1428,11 @@ EOF
   (##sys#check-number secs 'seconds->utc-time)
   (##sys#decode-seconds secs #t) )
 
-(define seconds->string 
+(define seconds->string
   (let ([ctime (foreign-lambda c-string "C_ctime" integer)])
-    (lambda (secs) 
+    (lambda (secs)
       (let ([str (ctime secs)])
-	(unless str (##sys#error 'seconds->string "can not convert seconds to string" secs))
+	(unless str (##sys#error 'seconds->string "cannot convert seconds to string" secs))
 	str) ) ) )
 
 (define time->string
@@ -1204,7 +1441,7 @@ EOF
       (##sys#check-vector tm 'time->string)
       (when (fx< (##sys#size tm) 10) (##sys#error 'time->string "time vector too short" tm))
       (let ([str (asctime tm)])
-	(unless str (##sys#error 'time->string "can not time vector to string" tm))
+	(unless str (##sys#error 'time->string "cannot time vector to string" tm))
 	str) ) ) )
 
 (define (local-time->seconds tm)
@@ -1212,7 +1449,7 @@ EOF
   (when (fx< (##sys#size tm) 10) (##sys#error 'local-time->seconds "time vector too short" tm))
   (if (##core#inline "C_mktime" tm)
       (##sys#cons-flonum)
-      (##sys#error 'local-time->seconds "can not convert time vector to seconds" tm) ) )
+      (##sys#error 'local-time->seconds "cannot convert time vector to seconds" tm) ) )
 
 
 
@@ -1240,10 +1477,10 @@ EOF
 		    [else (##sys#error 'set-buffering-mode! "invalid buffering-mode" mode port)] ) ] )
 	(##sys#check-exact size 'set-buffering-mode!)
 	(when (fx< (if (eq? 'stream (##sys#slot port 7))
-		       (##core#inline "C_setvbuf" port mode size) 
-		       -1) 
+		       (##core#inline "C_setvbuf" port mode size)
+		       -1)
 		   0)
-	  (##sys#error 'set-buffering-mode! "can not set buffering mode" port mode size) ) ) ) )
+	  (##sys#error 'set-buffering-mode! "cannot set buffering mode" port mode size) ) ) ) )
 
 ;;; Filename globbing:
 
@@ -1261,7 +1498,7 @@ EOF
 		(let ([rx (glob->regexp (make-pathname #f (or file "*") ext))])
 		  (let loop ([f (directory (or dir "."))])
 		    (cond [(null? f) (conc (cdr paths))]
-			  [(string-match rx (car f)) 
+			  [(string-match rx (car f))
 			   => (lambda (m) (cons (make-pathname dir (car m)) (loop (cdr f)))) ]
 			  [else (loop (cdr f))] ) ) ) ) ) ) ) ) ) )
 
@@ -1281,7 +1518,7 @@ EOF
 (define spawn/detach _p_detach)
 
 (define process-execute
-  (let ([setarg (foreign-lambda void "C_set_exec_arg" int scheme-pointer int)] 
+  (let ([setarg (foreign-lambda void "C_set_exec_arg" int scheme-pointer int)]
 	[freeargs (foreign-lambda void "C_free_exec_args")]
 	[pathname-strip-directory pathname-strip-directory] )
     (lambda (filename #!optional (arglist '()) envlist)
@@ -1292,19 +1529,19 @@ EOF
 	  (setarg 0 s (##sys#size s)) )
 	(do ([al arglist (cdr al)]
 	     [i 1 (fx+ i 1)] )
-	    ((null? al) 
+	    ((null? al)
 	     (setarg i #f 0)
 	     (let ([r (##core#inline "C_execvp" (##sys#make-c-string (##sys#expand-home-path filename)))])
 	       (##sys#update-errno)
 	       (when (fx= r -1)
 		 (freeargs)
-		 (##sys#error 'process-execute "can not execute process" filename) ) ) )
+		 (##sys#error 'process-execute "cannot execute process" filename) ) ) )
 	  (let ([s (car al)])
 	    (##sys#check-string s 'process-execute)
 	    (setarg i s (##sys#size s)) ) ) ) ) ) )
 
 (define process-spawn
-  (let ([setarg (foreign-lambda void "C_set_exec_arg" int scheme-pointer int)] 
+  (let ([setarg (foreign-lambda void "C_set_exec_arg" int scheme-pointer int)]
 	[freeargs (foreign-lambda void "C_free_exec_args")]
 	[pathname-strip-directory pathname-strip-directory] )
     (lambda (mode filename . arglist)
@@ -1316,22 +1553,22 @@ EOF
 	  (setarg 0 s (##sys#size s)) )
 	(do ([al arglist (cdr al)]
 	     [i 1 (fx+ i 1)] )
-	    ((null? al) 
+	    ((null? al)
 	     (setarg i #f 0)
 	     (let ([r (##core#inline "C_spawnvp" mode
 		       (##sys#make-c-string (##sys#expand-home-path filename)))])
 	       (##sys#update-errno)
 	       (when (fx= r -1)
 		 (freeargs)
-		 (##sys#error 'process-spawn "can not execute process" filename) )
-	       r) ) 
+		 (##sys#error 'process-spawn "cannot execute process" filename) )
+	       r) )
 	  (let ([s (car al)])
 	    (##sys#check-string s 'process-spawn)
 	    (setarg i s (##sys#size s)) ) ) ) ) ) )
 
 (define current-process-id (foreign-lambda int "C_getpid"))
 
-(define get-shell
+(define ##sys#shell-command
   (foreign-lambda* c-string () #<<EOF
     char *ret = getenv("COMSPEC");
     if (ret)
@@ -1348,13 +1585,17 @@ EOF
 EOF
     ) )
 
+(define (##sys#shell-command-arguments cmdlin)
+  (list "/c" cmdlin) )
+
 (define process-run
   (let ([process-spawn process-spawn]
 	[getenv getenv] )
     (lambda (f . args)
       (let ([args (if (pair? args) (car args) #f)])
-	(if args (process-spawn spawn/nowait f args)
-	    (process-spawn spawn/nowait (get-shell) (list "/c" f)) ) ) ) ) )
+	(if args
+	    (process-spawn spawn/nowait f args)
+	    (process-spawn spawn/nowait (##sys#shell-command) (##sys#shell-command-arguments f)) ) ) ) ) )
 
 ;;; Run subprocess connected with pipes:
 (define-foreign-variable _rdbuf char "C_rdbuf")
@@ -1364,56 +1605,82 @@ EOF
 (define close-handle
   (foreign-lambda int "close_handle" bool))
 
-(define process
-  (let ([make-input-port make-input-port]
-	[make-output-port make-output-port])
+; from original by Mejedi
+(define ##sys#process
+  (let (
+      [c-process
+        (foreign-lambda int "C_process" c-string c-pointer
+          (pointer int) (pointer int) (pointer int) (pointer int)
+          int)])
+    (lambda (loc cmd args env stdoutf stdinf stderrf)
+      (let (
+          [commandline
+            (if args
+              (let loop ([args args] [cmdlin cmd])
+                (if (null? args)
+                  cmdlin
+                  (loop (cdr args) (conc cmdlin " " (car args)))))
+              cmd)])
+        (let-location ([handle int -1] [stdin_fd int -1] [stdout_fd int -1] [stderr_fd int -1])
+          (let (
+              [code
+                (c-process commandline env
+                  (location handle) (location stdin_fd) (location stdout_fd) (location stderr_fd)
+                  (+ (if stdinf 0 1) (if stdoutf 0 2) (if stderrf 0 4)))])
+            (if (fx= 0 code)
+              (values
+                (and stdoutf (open-input-file* stdout_fd)) ;Parent stdin
+                (and stdinf (open-output-file* stdin_fd))  ;Parent stdout
+                handle
+                (and stderrf (open-input-file* stderr_fd)))
+              (begin
+                (##sys#update-errno)
+                (##sys#signal-hook #:process-error loc "cannot execute process" commandline))) ) ) ) ) ) )
+
+#;(define process (void))
+#;(define process* (void))
+(let ([%process
+        (lambda (loc err? cmd args env)
+          (##sys#check-string cmd loc)
+          (if args
+            (begin
+              (##sys#check-list args loc)
+              (for-each (cut ##sys#check-string <> loc) args) )
+              (begin
+                (set! args (##sys#shell-command-arguments cmd))
+                (set! cmd (##sys#shell-command)) ) )
+          (when env
+            (##sys#check-list env loc)
+            (for-each (cut ##sys#check-string <> loc) env) )
+          (receive [in out pid err]
+                      (##sys#process loc cmd args env #t #t err?)
+            (if err?
+              (values in out pid err)
+              (values in out pid) ) ) )] )
+  (set! process
     (lambda (cmd #!optional args env)
-      (if (##core#inline "C_redir_io")
-	(let* ([h0 _wr0]
-	       [h1 _rd1]
-	       [cmdline (string-append (get-shell) " /c " cmd "\x00")]
-	       [proc (##core#inline "C_run_process" cmdline)])
-	  (if (fx> proc 0)
-	    (values
-	      (make-input-port
-		(lambda () ; read
-		  (case (##core#inline "C_pipe_read" h1)
-		    [(-1) #!eof]
-		    [(0) (##sys#error 'process "could not read from pipe")]
-		    [(1) _rdbuf]) )
-		(lambda () ; ready?
-		  (##core#inline "C_pipe_ready" h1) )
-		(lambda () ; close
-		  (close-handle h1) )
-		;(lambda () ; peek
-		;)
-		)
-	      (make-output-port
-		(lambda (s) ; write
-		  (unless (##core#inline "C_pipe_write" h0 s (string-length s))
-		    (##sys#error 'process "could not write to pipe") ) )
-		(lambda () ; close
-		  (close-handle h0) )
-		)
-	      proc)
-	    (begin
-	      (##sys#update-errno)
-	      (C_close_handle h0)
-	      (C_close_handle h1)
-	      (##sys#error 'process "could not create process")))
-	  )
-	(##sys#error 'process "could not redirect I/O") ) ) ) )
+      (%process 'process #f cmd args env) ))
+  (set! process*
+    (lambda (cmd #!optional args env)
+      (%process 'process* #t cmd args env) )) )
 
 (define-foreign-variable _exstatus int "C_exstatus")
 
+(define (##sys#process-wait pid nohang)
+  (if (##core#inline "C_process_wait" pid nohang)
+    (values pid #t _exstatus)
+    (values -1 #f #f) ) )
+
 (define process-wait
   (lambda (pid . args)
     (let-optionals* args ([nohang #f])
-      (if (##core#inline "C_process_wait" pid nohang)
-	(values pid #t _exstatus)
-	(begin
+      (##sys#check-exact pid 'process-wait)
+      (receive [epid enorm ecode] (##sys#process-wait pid nohang)
+        (if (fx= epid -1)
+          (begin
 	    (##sys#update-errno)
-	    (values 0 #f 0) ) ) ) ) )
+	    (##sys#signal-hook #:process-error 'process-wait "waiting for child process failed" pid) )
+          (values epid enorm ecode) ) ) ) ) )
 
 (define sleep
   (lambda (t)
@@ -1429,7 +1696,7 @@ EOF
   (lambda ()
     (if (##core#inline "C_get_hostname")
       _hostname
-      (##sys#error 'get-host-name "can not retrieve host-name") ) ) )
+      (##sys#error 'get-host-name "cannot retrieve host-name") ) ) )
 
 (define system-information
   (lambda ()
@@ -1437,28 +1704,28 @@ EOF
       (list "windows" _hostname _osrel _osver _processor)
       (begin
 	(##sys#update-errno)
-	(##sys#error 'system-information "can not retrieve system-information") ) ) ) )
+	(##sys#error 'system-information "cannot retrieve system-information") ) ) ) )
 
 ;;; Find matching files:
 
-(define find-files 
+(define find-files
   (let ([glob glob]
 	[string-match string-match]
 	[make-pathname make-pathname]
 	[directory? directory?] )
     (lambda (dir pred . action-id-limit)
-      (let-optionals action-id-limit 
+      (let-optionals action-id-limit
 	  ([action (lambda (x y) (cons x y))] ; no eta reduction here - we want cons inlined.
-	   [id '()] 
+	   [id '()]
 	   [limit #f] )
 	(##sys#check-string dir 'find-files)
 	(let* ([depth 0]
 	       [lproc
 		(cond [(not limit) (lambda _ #t)]
 		      [(fixnum? limit) (lambda _ (fx< depth limit))]
-		      [else limit] ) ] 
-	       [pproc 
-		(if (string? pred) 
+		      [else limit] ) ]
+	       [pproc
+		(if (string? pred)
 		    (lambda (x) (string-match pred x))
 		    pred) ] )
 	  (let loop ([fs (glob (make-pathname dir "*"))]
@@ -1476,3 +1743,74 @@ EOF
 			       [else (loop rest r)] ) ]
 			[(pproc f) (loop rest (action f r))]
 			[else (loop rest r)] ) ) ) ) ) ) ) ) )
+
+;;; unimplemented stuff:
+
+(define-macro (define-unimplemented name)
+  `(define (,name . _)
+     (error ',name (##core#immutable '"this function is not available on this platform")) ) )
+
+(define-unimplemented change-file-owner)
+(define-unimplemented create-fifo)
+(define-unimplemented create-session)
+(define-unimplemented create-symbolic-link)
+(define-unimplemented current-effective-group-id)
+(define-unimplemented current-effective-user-id)
+(define-unimplemented current-group-id)
+(define-unimplemented current-user-id)
+(define-unimplemented map-file-to-memory)
+(define-unimplemented file-link)
+(define-unimplemented file-lock)
+(define-unimplemented file-lock/blocking)
+(define-unimplemented file-select)
+(define-unimplemented file-test-lock)
+(define-unimplemented file-truncate)
+(define-unimplemented file-unlock)
+(define-unimplemented get-groups)
+(define-unimplemented group-information)
+(define-unimplemented initialize-groups)
+(define-unimplemented local-timezone-abbreviation)
+(define-unimplemented memory-mapped-file-pointer)
+(define-unimplemented parent-process-id)
+(define-unimplemented process-fork)
+(define-unimplemented process-group-id)
+(define-unimplemented process-signal)
+(define-unimplemented read-symbolic-link)
+(define-unimplemented set-alarm!)
+(define-unimplemented set-group-id!)
+(define-unimplemented set-groups!)
+(define-unimplemented set-process-group-id!)
+(define-unimplemented set-root-directory!)
+(define-unimplemented set-signal-mask!)
+(define-unimplemented set-user-id!)
+(define-unimplemented signal-mask)
+(define-unimplemented signal-mask!)
+(define-unimplemented signal-masked?)
+(define-unimplemented signal-unmask!)
+(define-unimplemented terminal-name)
+(define-unimplemented terminal-port?)
+(define-unimplemented unmap-file-from-memory)
+(define-unimplemented user-information)
+(define-unimplemented utc-time->seconds)
+
+(define errno/wouldblock 0)
+
+(define (fifo? _) #f)
+(define (memory-mapped-file? _) #f)
+
+(define map/anonymous 0)
+(define map/file 0)
+(define map/fixed 0)
+(define map/private 0)
+(define map/shared 0)
+(define open/fsync 0)
+(define open/noctty 0)
+(define open/nonblock 0)
+(define open/sync 0)
+(define perm/isgid 0)
+(define perm/isuid 0)
+(define perm/isvtx 0)
+(define prot/exec 0)
+(define prot/none 0)
+(define prot/read 0)
+(define prot/write 0)
