@@ -1,6 +1,6 @@
-;;; srfi-18.scm - Simple thread unit - felix
+;;;; srfi-18.scm - Simple thread unit - felix
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -48,6 +48,7 @@
     (no-bound-checks)
     (no-procedure-checks-for-usual-bindings)
     (bound-to-procedure
+     ##sys#thread-yield!
      condition-property-accessor ##sys#tty-port? ##sys#thread-block-for-i/o thread-yield! ##sys#thread-unblock!
      ##sys#thread-basic-unblock! gensym ##sys#thread-block-for-timeout! ##sys#thread-kill!
      ##sys#thread-block-for-termination! make-thread ##sys#exact->inexact ##sys#flonum-fraction truncate
@@ -121,6 +122,11 @@ EOF
   (##sys#check-structure tm 'time 'time->seconds)
   (+ (##sys#slot tm 2) (/ (##sys#slot tm 3) 1000)) )
 
+(define (time->milliseconds tm)
+  (##sys#check-structure tm 'time 'time->milliseconds)
+  (+ (inexact->exact (* (- (##sys#slot tm 2) C_startup_time_seconds) 1000))
+     (##sys#slot tm 3) ) )
+
 (define (seconds->time n)
   (##sys#check-number n 'seconds->time)
   (let* ([n2 (max 0 (- n C_startup_time_seconds))] ; seconds since startup
@@ -128,6 +134,11 @@ EOF
 	 [n3 (inexact->exact (truncate (+ (* n2 1000) ms)))] ) ; milliseconds since startup
     (##sys#make-structure 'time n3 (truncate n) (inexact->exact ms)) ) )
 
+(define (milliseconds->time nms)
+  (##sys#check-exact nms 'milliseconds->time)
+  (let ((s (+ C_startup_time_seconds (/ nms 1000))))
+    (##sys#make-structure 'time nms s 0) ) )
+
 (define (time? x) (##sys#structure? x 'time))
 
 (define srfi-18:time? time?)
@@ -219,12 +230,7 @@ EOF
       (##sys#add-to-ready-queue thread) 
       thread) ) )
 
-(define (thread-yield!)
-  (##sys#call-with-current-continuation
-   (lambda (return)
-     (let ((ct ##sys#current-thread))
-       (##sys#setslot ct 1 (lambda () (return (##core#undefined))))
-       (##sys#schedule) ) ) ) )
+(define thread-yield! ##sys#thread-yield!) ;In library.scm
 
 (define thread-join!
   (lambda (thread . timeout)
@@ -281,16 +287,16 @@ EOF
     (##sys#setslot thread 3 'ready)
     (##sys#add-to-ready-queue thread) ) )
 
-(define thread-sleep!
-  (lambda (tm)
-    (unless tm (##sys#signal-hook #:type-error 'thread-sleep! "invalid timeout argument" tm))
+(define (thread-sleep! tm)
+  (define (sleep limit loc)
     (##sys#call-with-current-continuation
      (lambda (return)
-       (let ([limit (##sys#compute-time-limit tm)]
-	     [ct ##sys#current-thread] )
+       (let ((ct ##sys#current-thread))
 	 (##sys#setslot ct 1 (lambda () (return (##core#undefined))))
 	 (##sys#thread-block-for-timeout! ct limit)
-	 (##sys#schedule) ) ) ) ) )
+	 (##sys#schedule) ) ) ) )
+  (unless tm (##sys#signal-hook #:type-error 'thread-sleep! "invalid timeout argument" tm))
+  (sleep (##sys#compute-time-limit tm) 'thread-sleep!) )
 
 
 ;;; Mutexes:
