@@ -1,6 +1,6 @@
 ;;;; srfi-4.scm - Homogeneous numeric vectors
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -64,8 +64,8 @@
 #define C_s32poke(b, i, x)     ((((C_s32 *)C_data_pointer(b))[ C_unfix(i) ] = C_num_to_int(x)), C_SCHEME_UNDEFINED)
 #define C_f32poke(b, i, x)     ((((float *)C_data_pointer(b))[ C_unfix(i) ] = C_flonum_magnitude(x)), C_SCHEME_UNDEFINED)
 #define C_f64poke(b, i, x)     ((((double *)C_data_pointer(b))[ C_unfix(i) ] = C_flonum_magnitude(x)), C_SCHEME_UNDEFINED)
-#define C_copy_subvector(to, from, start_from, bytes)   \
-  (C_memcpy((C_char *)C_data_pointer(to), (C_char *)C_data_pointer(from) + C_unfix(start_from), C_unfix(bytes)), \
+#define C_copy_subvector(to, from, start_to, start_from, bytes)   \
+  (C_memcpy((C_char *)C_data_pointer(to) + C_unfix(start_to), (C_char *)C_data_pointer(from) + C_unfix(start_from), C_unfix(bytes)), \
     C_SCHEME_UNDEFINED)
 EOF
 ) )
@@ -510,7 +510,8 @@ EOF
     (lambda (str)
       (##sys#check-byte-vector str loc)
       (let ([len (##core#inline "C_block_size" str)])
-	(if (or (eq? #t sz) (##core#inline "C_eqp" 0 (fxmod len sz)))
+	(if (or (eq? #t sz) 
+		(##core#inline "C_eqp" 0 (##core#inline "C_fixnum_modulo" len sz)))
 	    (##sys#make-structure tag str) 
 	    (##sys#error loc "bytevector does not have correct size for packing" tag len sz) ) ) ) )
 
@@ -583,14 +584,14 @@ EOF
   (##sys#check-structure v t loc)
   (let* ([bv (##sys#slot v 1)]
 	 [len (##sys#size bv)]
-	 [ilen (fx/ len es)] )
+	 [ilen (##core#inline "C_fixnum_divide" len es)] )
     (##sys#check-range from 0 (fx+ ilen 1) loc)
     (##sys#check-range to 0 (fx+ ilen 1) loc)
     (let* ([size2 (fx* es (fx- to from))]
 	   [bv2 (##sys#allocate-vector size2 #t #f #t)] )
       (##core#inline "C_string_to_bytevector" bv2)
       (let ([v (##sys#make-structure t bv2)])
-	(##core#inline "C_copy_subvector" bv2 bv (fx* from es) size2)
+	(##core#inline "C_copy_subvector" bv2 bv 0 (fx* from es) size2)
 	v) ) ) )
 
 (define (subu8vector v from to) (subvector v 'u8vector 1 from to 'subu8vector))
@@ -602,5 +603,49 @@ EOF
 (define (subf32vector v from to) (subvector v 'f32vector 4 from to 'subf32vector))
 (define (subf64vector v from to) (subvector v 'f64vector 8 from to 'subf64vector))
 
+(define (write-u8vector v #!optional (port ##sys#standard-output) (from 0) (to (u8vector-length v)))
+  (##sys#check-structure v 'u8vector 'write-u8vector)
+  (##sys#check-port port 'write-u8vector)
+  (let ((buf (##sys#slot v 1)))
+    (do ((i from (fx+ i 1)))
+	((fx>= i to))
+      (##sys#write-char-0 (integer->char (##core#inline "C_u8peek" buf i)) port) ) ) )
+
+(define (read-u8vector! n dest #!optional (port ##sys#standard-input) (start 0))
+  (##sys#check-port port 'read-u8vector!)
+  (##sys#check-exact start 'read-u8vector!)
+  (##sys#check-structure dest 'u8vector 'read-u8vector!)
+  (let ((dest (##sys#slot dest 1)))
+    (when n
+      (##sys#check-exact n 'read-u8vector!)
+      (when (fx> (fx+ start n) (##sys#size dest))
+	(set! n (fx- (##sys#size dest) start))))
+    (##sys#read-string! n dest port start) ) )
+
+(define read-u8vector
+  (let ((open-output-string open-output-string)
+	(get-output-string get-output-string) )
+    (lambda (#!optional n (p ##sys#standard-input))
+      (##sys#check-port p 'read-u8vector)
+      (cond (n (##sys#check-exact n 'read-u8vector)
+	       (let* ((str (##sys#allocate-vector n #t #f #t))
+		      (n2 (##sys#read-string! n str p 0)) )
+		 (##sys#make-structure
+		  'u8vector
+		  (if (eq? n n2)
+		      str
+		      (let ((str2 (##sys#allocate-vector n2 #t #f #t)))
+			(##core#inline "C_substring_copy" str str2 0 n2 0)
+			str2) ) ) ) )
+	    (else
+	     (let ([str (open-output-string)])
+	       (let loop ([n n])
+		 (or (and (eq? n 0) (get-output-string str))
+		     (let ([c (##sys#read-char-0 p)])
+		       (if (eof-object? c)
+			   (get-output-string str)
+			   (begin
+			     (##sys#write-char c str) 
+			     (loop (and n (fx- n 1))) ) ) ) ) ) ) ) ) ) ) )
 
 (register-feature! 'srfi-4)
