@@ -1,6 +1,6 @@
 ;;;; support.scm - Miscellaneous support code for the CHICKEN compiler
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -49,7 +49,7 @@
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
-  dependency-list broken-constant-nodes inline-substitutions-enabled
+  dependency-list broken-constant-nodes inline-substitutions-enabled emit-syntax-trace-info
   always-bound-to-procedure block-variable-literal copy-node! valid-c-identifier? tree-copy copy-node-tree-and-rename
   direct-call-ids foreign-type-table first-analysis scan-sharp-greater-string
   expand-profile-lambda profile-lambda-list profile-lambda-index profile-info-vector-name
@@ -74,7 +74,7 @@
   default-optimization-iterations chop-separator chop-extension follow-without-loop dump-exported-globals
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
   foreign-argument-conversion foreign-result-conversion final-foreign-type debugging export-list block-globals
-  lookup-exports-file constant-declarations
+  lookup-exports-file constant-declarations process-lambda-documentation
   make-random-name foreign-type-convert-result foreign-type-convert-argument process-custom-declaration}
 
 
@@ -121,10 +121,16 @@
 (set! ##sys#syntax-error-hook
   (lambda (msg . args)
     (let ([out (current-error-port)])
-      (fprintf out "Syntax error: ~a~%" msg) 
-      (for-each (lambda (x) (write x out) (newline out)) args)
+      (fprintf out "Syntax error: ~a~%~%" msg) 
+      (for-each (cut fprintf out "\t~s~%" <>) args)
+      (print-call-chain out 0 ##sys#current-thread "\n\tExpansion history:\n")
       (exit 70) ) ) )
 
+(set! syntax-error ##sys#syntax-error-hook)
+
+(define (emit-syntax-trace-info info cntr) 
+  (##core#inline "C_emit_syntax_trace_info" info cntr ##sys#current-thread) )
+
 (define (map-llist proc llist)
   (let loop ([llist llist])
     (cond [(null? llist) '()]
@@ -335,6 +341,9 @@
 
 (define decompose-lambda-list ##sys#decompose-lambda-list)
 
+(define (process-lambda-documentation id doc)
+  #f)					; Hook this
+
 
 ;;; Profiling instrumentation:
 
@@ -440,7 +449,7 @@
   (let ((names '((captured . cpt) (assigned . set) (boxed . box) (global . glo) (assigned-locally . stl)
 		 (contractable . con) (standard-binding . stb) (foldable . fld) (simple . sim) (inlinable . inl)
 		 (side-effecting . sef) (collapsable . col) (removable . rem) (constant . con)
-		 (undefined . und) (replacing . rpg) (unused . uud) (extended-binding . xtb)
+		 (undefined . und) (replacing . rpg) (unused . uud) (extended-binding . xtb) (inline-export . ilx)
 		 (customizable . cst) (has-unused-parameters . hup) (boxed-rest . bxr) ) ) )
     (lambda (db)
       (##sys#hash-table-for-each
@@ -455,7 +464,7 @@
 		 (begin
 		   (case (caar es)
 		     ((captured assigned boxed global contractable standard-binding foldable assigned-locally
-		       side-effecting collapsable removable undefined replacing unused simple inlinable
+		       side-effecting collapsable removable undefined replacing unused simple inlinable inline-export
 		       has-unused-parameters extended-binding customizable constant boxed-rest)
 		      (printf "\t~a" (cdr (assq (caar es) names))) )
 		     ((unknown)
@@ -752,15 +761,22 @@
   (unless block-compilation
     (with-output-to-file file
       (lambda ()
-	(##sys#hash-table-for-each
-	 (lambda (sym plist)
-	   (when (and (assq 'global plist) 
-		      (assq 'assigned plist)
-		      (or (and export-list (memq sym export-list))
-			  (not (memq sym block-globals)) ) )
-	     (write sym) 
-	     (newline) ) )
-	 db) ) ) ) )
+	(let ((exports '()))
+	  (##sys#hash-table-for-each
+	   (lambda (sym plist)
+	     (when (and (assq 'global plist) 
+			(assq 'assigned plist)
+			(or (and export-list (memq sym export-list))
+			    (not (memq sym block-globals)) ) )
+	       (set! exports (cons sym exports)) ) )
+	   db)
+	  (for-each 
+	   (lambda (s)
+	     (write s)
+	     (newline) )
+	   (sort exports
+		 (lambda (s1 s2)
+		   (string<? (##sys#slot s1 1) (##sys#slot s2 1)))) ) ) ) ) ) )
 
 (define (dump-undefined-globals db)
   (##sys#hash-table-for-each
