@@ -1,6 +1,6 @@
 ;;;; tcp.scm - Networking stuff
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -340,27 +340,29 @@ EOF
 	     (oclosed #f)
 	     (outbufsize (tbs))
 	     (outbuf (and outbufsize (fx> outbufsize 0) ""))
+	     (read-input
+	      (lambda ()
+		(let loop ()
+		  (let ((n (##net#recv fd buf +input-buffer-size+ 0)))
+		    (cond ((eq? -1 n)
+			   (cond ((eq? errno _ewouldblock) 
+				  (##sys#thread-block-for-i/o! ##sys#current-thread fd #t)
+				  (yield)
+				  (loop) )
+				 (else
+				  (##sys#update-errno)
+				  (##sys#signal-hook 
+				   #:network-error
+				   (##sys#string-append "can not read from socket - " strerror) 
+				   fd) ) ) )
+			  (else
+			   (set! buflen n)
+			   (set! bufindex 0) ) ) ) ) ) )
 	     (in
 	      (make-input-port
 	       (lambda ()
 		 (when (fx>= bufindex buflen)
-		   (let ((n (let loop ()
-			      (let ((n (##net#recv fd buf +input-buffer-size+ 0)))
-				(if (eq? -1 n)
-				    (if (eq? errno _ewouldblock) 
-					(begin
-					  (##sys#thread-block-for-i/o! ##sys#current-thread fd #t)
-					  (yield)
-					  (loop) )
-					(begin
-					  (##sys#update-errno)
-					  (##sys#signal-hook 
-					   #:network-error
-					   (##sys#string-append "can not read from socket - " strerror) 
-					   fd) ) )
-				    n) ) ) ) )
-		     (set! buflen n)
-		     (set! bufindex 0) ) )
+		   (read-input))
 		 (if (fx>= bufindex buflen)
 		     #!eof
 		     (let ((c (##core#inline "C_subchar" buf bufindex)))
@@ -371,8 +373,10 @@ EOF
 		     (let ((f (##net#select fd)))
 		       (when (eq? f -1)
 			 (##sys#update-errno)
-			 (##sys#signal-hook #:network-error (##sys#string-append "can not check socket for input - " strerror) 
-					    fd) )
+			 (##sys#signal-hook
+			  #:network-error
+			  (##sys#string-append "can not check socket for input - " strerror) 
+			  fd) )
 		       (eq? f 1) ) ) )
 	       (lambda ()
 		 (unless iclosed
@@ -380,8 +384,49 @@ EOF
 		   (unless (##sys#slot data 1) (##net#shutdown fd _sd_receive))
 		   (when (and oclosed (eq? -1 (##net#close fd)))
 		     (##sys#update-errno)
-		     (##sys#signal-hook #:network-error (##sys#string-append "can not close socket input port - " strerror)
-					fd) ) ) ) ) )
+		     (##sys#signal-hook
+		      #:network-error
+		      (##sys#string-append "can not close socket input port - " strerror)
+		      fd) ) ) )
+	       #f
+	       (lambda (p n dest start)	; read-string!
+		 (let loop ((n n) (m 0) (start start))
+		   (cond ((eq? n 0) m)
+			 ((fx< bufindex buflen)
+			  (let* ((rest (fx- buflen bufindex))
+				 (n2 (if (fx< n rest) n rest)))
+			    (##core#inline "C_substring_copy" buf dest bufindex (fx+ bufindex n2) start)
+			    (set! bufindex (fx+ bufindex n2))
+			    (loop (fx- n n2) (fx+ m n2) (fx+ start n2)) ) )
+			 (else
+			  (read-input)
+			  (if (eq? buflen 0) 
+			      m
+			      (loop n m start) ) ) ) ) )
+	       (lambda (p limit)	; read-line
+		 (let loop ((str #f))
+		   (cond ((fx< bufindex buflen)
+			  (##sys#scan-buffer-line
+			   buf 
+			   buflen
+			   bufindex
+			   (lambda (pos2 next)
+			     (let ((dest (##sys#make-string (fx- pos2 bufindex))))
+			       (##core#inline "C_substring_copy" buf dest bufindex pos2 0)
+			       (set! bufindex next)
+			       (cond ((eq? pos2 next) ; no line-terminator encountered
+				      (read-input)
+				      (if (fx>= bufindex buflen)
+					  (or str "")
+					  (loop (if str (##sys#string-append str dest) dest)) ) )
+				     (else 
+				      (##sys#setislot p 4 (fx+ (##sys#slot p 4) 1))
+				      (if str (##sys#string-append str dest) dest)) ) ) ) ) )
+			 (else
+			  (read-input)
+			  (if (fx< bufindex buflen)
+			      (loop str)
+			      #!eof) ) ) ) ) ) )
 	     (output
 	      (lambda (s)
 		(let loop ((len (##sys#size s)) 
